# This file is part of allegedb, an object relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""allegedb's special implementations of the NetworkX graph objects"""

from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import MutableMapping
from itertools import chain
from typing import TYPE_CHECKING

import networkx
from networkx.exception import NetworkXError

from .types import Branch, Key, Tick, Turn
from .wrap import MutableMappingUnwrapper, SpecialMapping, wrapval

if TYPE_CHECKING:
	from .engine import Engine


class EntityCollisionError(ValueError):
	"""For when there's a discrepancy between the kind of entity you're creating and the one by the same name"""


def getatt(attribute_name):
	"""An easy way to make an alias"""
	from operator import attrgetter

	ret = property(attrgetter(attribute_name))
	ret.__doc__ = "Alias to `{}`".format(attribute_name)
	return ret


_alleged_receivers = defaultdict(list)


class AllegedMapping(MutableMappingUnwrapper, SpecialMapping, ABC):
	"""Common amenities for mappings"""

	__slots__ = ()

	def clear(self):
		"""Delete everything"""
		for k in list(self.keys()):
			if k in self:
				del self[k]


class AbstractEntityMapping(AllegedMapping, ABC):
	__slots__ = ()
	db: "Engine"

	@abstractmethod
	def _get_cache(
		self, key: Key, branch: Branch, turn: Turn, tick: Tick
	) -> dict:
		raise NotImplementedError

	def _get_cache_now(self, key):
		return self._get_cache(key, *self.db._btt())

	@abstractmethod
	def _cache_contains(self, key, branch, turn, tick):
		raise NotImplementedError

	@abstractmethod
	def _set_db(self, key, branch, turn, tick, value):
		"""Set a value for a key in the database (not the cache)."""
		raise NotImplementedError

	@abstractmethod
	def _set_cache(self, key, branch, turn, tick, value):
		raise NotImplementedError

	def _del_db(self, key, branch, turn, tick):
		"""Delete a key from the database (not the cache)."""
		self._set_db(key, branch, turn, tick, ...)

	def _del_cache(self, key, branch, turn, tick):
		self._set_cache(key, branch, turn, tick, ...)

	def __getitem__(self, key):
		"""If key is 'graph', return myself as a dict, else get the present
		value of the key and return that

		"""

		return wrapval(self, key, self._get_cache_now(key))

	def __contains__(self, item):
		return item == "name" or self._cache_contains(item, *self.db._btt())

	def __setitem__(self, key, value):
		"""Set key=value at the present branch and revision"""
		if value is ...:
			raise ValueError(
				"Lisien uses the ellipsis to indicate that a key's been deleted"
			)
		branch, turn, tick = self.db._nbtt()
		try:
			if self._get_cache(key, branch, turn, tick) != value:
				self._set_cache(key, branch, turn, tick, value)
		except KeyError:
			self._set_cache(key, branch, turn, tick, value)
		self._set_db(key, branch, turn, tick, value)

	def __delitem__(self, key):
		branch, turn, tick = self.db._nbtt()
		self._del_cache(key, branch, turn, tick)
		self._del_db(key, branch, turn, tick)


class GraphMapping(AbstractEntityMapping):
	"""Mapping for graph attributes"""

	__slots__ = (
		"graph",
		"db",
		"_iter_stuff",
		"_cache_contains_stuff",
		"_len_stuff",
		"_get_stuff",
		"_set_db_stuff",
		"_set_cache_stuff",
		"_del_db_stuff",
		"_get_cache_stuff",
	)

	def __init__(self, graph):
		super().__init__(graph)
		self.graph = graph
		self.db = db = graph.db
		btt = db._btt
		graph_val_cache = db._graph_val_cache
		graphn = graph.name
		self._iter_stuff = (graph_val_cache.iter_keys, graphn, btt)
		self._cache_contains_stuff = (graph_val_cache.contains_key, graphn)
		self._len_stuff = (graph_val_cache.count_keys, graphn, btt)
		self._get_stuff = (self._get_cache, btt)
		graph_val_set = db.query.graph_val_set
		self._set_db_stuff = (graph_val_set, graphn)
		self._set_cache_stuff = (graph_val_cache.store, graphn)
		self._del_db_stuff = (graph_val_set, graphn)
		self._get_cache_stuff = (graph_val_cache.retrieve, graphn)

	def __iter__(self):
		iter_entity_keys, graphn, btt = self._iter_stuff
		yield "name"
		yield from iter_entity_keys(graphn, *btt())

	def _cache_contains(self, key, branch, turn, tick):
		contains_key, graphn = self._cache_contains_stuff
		return contains_key(graphn, key, branch, turn, tick)

	def __len__(self):
		count_keys, graphn, btt = self._len_stuff
		return 1 + count_keys(graphn, *btt())

	def __getitem__(self, item):
		if item == "name":
			return self.graph.name
		return super().__getitem__(item)

	def __setitem__(self, key, value):
		if key == "name":
			raise KeyError("name cannot be changed after creation")
		super().__setitem__(key, value)

	def _get_cache(self, key, branch, turn, tick):
		retrieve, graphn = self._get_cache_stuff
		return retrieve(graphn, key, branch, turn, tick)

	def _get(self, key):
		get_cache, btt = self._get_stuff
		return get_cache(key, *btt())

	def _set_db(self, key, branch, turn, tick, value):
		graph_val_set, graphn = self._set_db_stuff
		graph_val_set(graphn, key, branch, turn, tick, value)

	def _set_cache(self, key, branch, turn, tick, value):
		store, graphn = self._set_cache_stuff
		store(graphn, key, branch, turn, tick, value)

	def _del_db(self, key, branch, turn, tick):
		graph_val_set, graphn = self._del_db_stuff
		graph_val_set(graphn, key, branch, turn, tick, ...)

	def clear(self):
		keys = set(self.keys())
		keys.remove("name")
		for k in keys:
			del self[k]

	def unwrap(self):
		return {
			k: v.unwrap()
			if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap")
			else v
			for (k, v) in self.items()
		}

	def __eq__(self, other):
		if hasattr(other, "unwrap"):
			other = other.unwrap()
		other = other.copy()
		me = self.unwrap().copy()
		if "name" not in other:
			del me["name"]
		return me == other


class Node(AbstractEntityMapping):
	"""Mapping for node attributes"""

	__slots__ = (
		"graph",
		"name",
		"db",
		"_iter_stuff",
		"_cache_contains_stuff",
		"_len_stuff",
		"_get_cache_stuff",
		"_set_db_stuff",
		"_set_cache_stuff",
	)

	def _validate_node_type(self):
		return True

	def __init__(self, graph, node):
		"""Store name and graph"""
		super().__init__(graph)
		self.graph = graph
		self.name = node
		self.db = db = graph.db
		node_val_cache = db._node_val_cache
		graphn = graph.name
		btt = db._btt
		self._iter_stuff = (
			node_val_cache.iter_keys,
			graphn,
			node,
			btt,
		)
		self._cache_contains_stuff = (
			node_val_cache.contains_key,
			graphn,
			node,
		)
		self._len_stuff = (
			node_val_cache.count_keys,
			graphn,
			node,
			btt,
		)
		self._get_cache_stuff = (node_val_cache.retrieve, graphn, node)
		self._set_db_stuff = (db.query.node_val_set, graphn, node)
		self._set_cache_stuff = (db._node_val_cache.store, graphn, node)

	def __repr__(self):
		return "{}(graph={}, name={})".format(
			self.__class__.__name__, repr(self.graph), repr(self.name)
		)

	def __str__(self):
		return f"Node of class {self.__class__.__name__} named {self.name}with stats {dict(self)}"

	def __iter__(self):
		iter_entity_keys, graphn, node, btt = self._iter_stuff
		return iter_entity_keys(graphn, node, *btt())

	def _cache_contains(self, key, branch, turn, tick):
		contains_key, graphn, node = self._cache_contains_stuff
		return contains_key(graphn, node, key, branch, turn, tick)

	def __len__(self):
		count_entity_keys, graphn, node, btt = self._len_stuff
		return count_entity_keys(graphn, node, *btt())

	def _get_cache(self, key, branch, turn, tick):
		retrieve, graphn, node = self._get_cache_stuff
		return retrieve(graphn, node, key, branch, turn, tick)

	def _set_db(self, key, branch, turn, tick, value):
		node_val_set, graphn, node = self._set_db_stuff
		node_val_set(graphn, node, key, branch, turn, tick, value)

	def _set_cache(self, key, branch, turn, tick, value):
		store, graphn, node = self._set_cache_stuff
		store(graphn, node, key, branch, turn, tick, value)

	def __eq__(self, other):
		if not hasattr(other, "keys") or not callable(other.keys):
			return False
		if not hasattr(other, "name"):
			return False
		if self.name != other.name:
			return False
		if not hasattr(other, "graph"):
			return False
		if self.graph.name != other.graph.name:
			return False
		if self.keys() != other.keys():
			return False
		for key in self:
			if self[key] != other[key]:
				return False
		return True


class Edge(AbstractEntityMapping):
	"""Mapping for edge attributes"""

	__slots__ = (
		"graph",
		"orig",
		"dest",
		"db",
		"_iter_stuff",
		"_cache_contains_stuff",
		"_len_stuff",
		"_get_cache_stuff",
		"_set_db_stuff",
		"_set_cache_stuff",
	)

	set_db_time = set_cache_time = 0

	def __init__(self, graph, orig, dest):
		super().__init__(graph)
		self.graph = graph
		self.db = db = graph.db
		self.orig = orig
		self.dest = dest
		edge_val_cache = db._edge_val_cache
		graphn = graph.name
		btt = db._btt
		self._iter_stuff = (
			edge_val_cache.iter_keys,
			graphn,
			orig,
			dest,
			btt,
		)
		self._cache_contains_stuff = (
			edge_val_cache.contains_key,
			graphn,
			orig,
			dest,
		)
		self._len_stuff = (
			edge_val_cache.count_keys,
			graphn,
			orig,
			dest,
			btt,
		)
		self._get_cache_stuff = (
			edge_val_cache.retrieve,
			graphn,
			orig,
			dest,
		)
		self._set_db_stuff = (db.query.edge_val_set, graphn, orig, dest)
		self._set_cache_stuff = (edge_val_cache.store, graphn, orig, dest)

	def __repr__(self):
		return "<{} in graph {} from {} to {} containing {}>".format(
			self.__class__.__name__,
			self.graph.name,
			self.orig,
			self.dest,
			dict(self),
		)

	def __str__(self):
		return str(dict(self))

	def __iter__(self):
		iter_entity_keys, graphn, orig, dest, btt = self._iter_stuff
		return iter_entity_keys(graphn, orig, dest, *btt())

	def _cache_contains(self, key, branch, turn, tick):
		contains_key, graphn, orig, dest = self._cache_contains_stuff
		return contains_key(graphn, orig, dest, key, branch, turn, tick)

	def __len__(self):
		count_entity_keys, graphn, orig, dest, btt = self._len_stuff
		return count_entity_keys(graphn, orig, dest, *btt())

	def _get_cache(self, key, branch, turn, tick):
		retrieve, graphn, orig, dest = self._get_cache_stuff
		return retrieve(graphn, orig, dest, key, branch, turn, tick)

	def _set_db(self, key, branch, turn, tick, value):
		edge_val_set, graphn, orig, dest = self._set_db_stuff
		edge_val_set(graphn, orig, dest, key, branch, turn, tick, value)

	def _set_cache(self, key, branch, turn, tick, value):
		store, graphn, orig, dest = self._set_cache_stuff
		store(graphn, orig, dest, key, branch, turn, tick, value)


class GraphNodeMapping(AllegedMapping):
	"""Mapping for nodes in a graph"""

	__slots__ = ("graph",)

	db = getatt("graph.db")
	"""Alias to ``self.graph.db``"""

	def __init__(self, graph):
		super().__init__(graph)
		self.graph = graph

	def __iter__(self):
		"""Iterate over the names of the nodes"""
		now = self.db._btt()
		gn = self.graph.name
		nc = self.db._nodes_cache
		for entity in nc.iter_entities(gn, *now):
			if entity in self:
				yield entity

	def __eq__(self, other):
		from collections.abc import Mapping

		if not isinstance(other, Mapping):
			return NotImplemented
		if self.keys() != other.keys():
			return False
		for k in self.keys():
			me = self[k]
			you = other[k]
			if hasattr(me, "unwrap") and not hasattr(me, "no_unwrap"):
				me = me.unwrap()
			if hasattr(you, "unwrap") and not hasattr(you, "no_unwrap"):
				you = you.unwrap()
			if me != you:
				return False
		else:
			return True

	def __contains__(self, node):
		"""Return whether the node exists presently"""
		return self.db._nodes_cache.contains_entity(
			self.graph.name, node, *self.db._btt()
		)

	def __len__(self):
		"""How many nodes exist right now?"""
		return self.db._nodes_cache.count_entities(
			self.graph.name, *self.db._btt()
		)

	def __getitem__(self, node):
		"""If the node exists at present, return it, else throw KeyError"""
		if node not in self:
			raise KeyError
		return self.db._get_node(self.graph, node)

	def __setitem__(self, node, dikt):
		"""Only accept dict-like values for assignment. These are taken to be
		dicts of node attributes, and so, a new GraphNodeMapping.Node
		is made with them, perhaps clearing out the one already there.

		"""
		created = False
		db = self.db
		graph = self.graph
		gname = graph.name
		if not db._node_exists(gname, node):
			created = True
			db._exist_node(gname, node, True)
		n = db._get_node(graph, node)
		n.clear()
		n.update(dikt)

	def __delitem__(self, node):
		"""Indicate that the given node no longer exists"""
		if node not in self:
			raise KeyError("No such node")
		for succ in self.graph.adj[node]:
			del self.graph.adj[node][succ]
		for pred in self.graph.pred[node]:
			del self.graph.pred[node][pred]
		branch, turn, tick = self.db._nbtt()
		self.db.query.exist_node(
			self.graph.name, node, branch, turn, tick, False
		)
		self.db._nodes_cache.store(
			self.graph.name, node, branch, turn, tick, False
		)
		key = (self.graph.name, node)
		if node in self.db._node_objs:
			del self.db._node_objs[key]

	def update(self, m, /, **kwargs):
		for node, value in chain(m.items(), kwargs.items()):
			if value is ...:
				del self[node]
			elif node not in self:
				self[node] = value
			else:
				self[node].update(value)


class GraphEdgeMapping(AllegedMapping):
	"""Provides an adjacency mapping and possibly a predecessor mapping
	for a graph.

	"""

	__slots__ = ("graph", "_cache")

	db = getatt("graph.db")
	"""Alias to ``self.graph.db``"""

	def __init__(self, graph):
		super().__init__(graph)
		self.graph = graph
		self._cache = {}

	def __eq__(self, other):
		"""Compare dictified versions of the edge mappings within me.

		As I serve custom Predecessor or Successor classes, which
		themselves serve the custom Edge class, I wouldn't normally be
		comparable to a networkx adjacency dictionary. Converting
		myself and the other argument to dicts allows the comparison
		to work anyway.

		"""
		if not hasattr(other, "keys"):
			return False
		if self.keys() != other.keys():
			return False
		for k in self.keys():
			if dict(self[k]) != dict(other[k]):
				return False
		return True

	def __iter__(self):
		return iter(self.graph.node)


class AbstractSuccessors(GraphEdgeMapping):
	__slots__ = ("graph", "container", "orig", "_cache")

	db = getatt("graph.db")
	"""Alias to ``self.graph.db``"""

	def _order_nodes(self, node):
		raise NotImplementedError

	def __init__(self, container, orig):
		"""Store container and node"""
		super().__init__(container.graph)
		self.container = container
		self.orig = orig

	def __iter__(self):
		"""Iterate over node IDs that have an edge with my orig"""
		for that in self.db._edges_cache.iter_successors(
			self.graph.name, self.orig, *self.db._btt()
		):
			if that in self:
				yield that

	def __contains__(self, dest):
		"""Is there an edge leading to ``dest`` at the moment?"""
		orig, dest = self._order_nodes(dest)
		return self.db._edges_cache.has_successor(
			self.graph.name, orig, dest, *self.db._btt()
		)

	def __len__(self):
		"""How many nodes touch an edge shared with my orig?"""
		n = 0
		for n, _ in enumerate(self, start=1):
			pass
		return n

	def _make_edge(self, dest):
		return Edge(self.graph, *self._order_nodes(dest))

	def __getitem__(self, dest):
		"""Get the edge between my orig and the given node"""
		if dest not in self:
			raise KeyError("No edge {}->{}".format(self.orig, dest))
		orig, dest = self._order_nodes(dest)
		return self.db._get_edge(self.graph, orig, dest)

	def __setitem__(self, dest, value):
		"""Set the edge between my orig and the given dest to the given
		value, a mapping.

		"""
		real_dest = dest
		orig, dest = self._order_nodes(dest)
		created = dest not in self
		if orig not in self.graph.node:
			self.graph.add_node(orig)
		if dest not in self.graph.node:
			self.graph.add_node(dest)
		branch, turn, tick = self.db._nbtt()
		self.db.query.exist_edge(
			self.graph.name, orig, dest, 0, branch, turn, tick, True
		)
		self.db._edges_cache.store(
			self.graph.name, orig, dest, branch, turn, tick, True
		)
		e = self[real_dest]
		e.clear()
		e.update(value)

	def __delitem__(self, dest):
		"""Remove the edge between my orig and the given dest"""
		branch, turn, tick = self.db._nbtt()
		orig, dest = self._order_nodes(dest)
		self.db.query.exist_edge(
			self.graph.name, orig, dest, 0, branch, turn, tick, False
		)
		self.db._edges_cache.store(
			self.graph.name, orig, dest, branch, turn, tick, None
		)

	def __repr__(self):
		cls = self.__class__
		return "<{}.{} object containing {}>".format(
			cls.__module__, cls.__name__, dict(self)
		)

	def clear(self):
		"""Delete every edge with origin at my orig"""
		for dest in list(self):
			del self[dest]


class GraphSuccessorsMapping(GraphEdgeMapping):
	"""Mapping for Successors (itself a MutableMapping)"""

	__slots__ = ("graph",)

	class Successors(AbstractSuccessors):
		__slots__ = ("graph", "container", "orig", "_cache")

		def _order_nodes(self, dest):
			if dest < self.orig:
				return (dest, self.orig)
			else:
				return (self.orig, dest)

	def __getitem__(self, orig):
		if orig not in self._cache:
			self._cache[orig] = self.Successors(self, orig)
		return self._cache[orig]

	def __setitem__(self, key, val):
		"""Wipe out any edges presently emanating from orig and replace them
		with those described by val

		"""
		if key in self:
			sucs = self[key]
			sucs.clear()
		else:
			sucs = self._cache[key] = self.Successors(self, key)
		if val:
			sucs.update(val)

	def __delitem__(self, key):
		"""Wipe out edges emanating from orig"""
		self[key].clear()
		del self._cache[key]

	def __iter__(self):
		for node in self.graph.node:
			if node in self:
				yield node

	def __len__(self):
		n = 0
		for node in self.graph.node:
			if node in self:
				n += 1
		return n

	def __contains__(self, key):
		return key in self.graph.node

	def __repr__(self):
		cls = self.__class__
		return "<{}.{} object containing {}>".format(
			cls.__module__,
			cls.__name__,
			{
				k: {k2: dict(v2) for (k2, v2) in v.items()}
				for (k, v) in self.items()
			},
		)


class DiGraphSuccessorsMapping(GraphSuccessorsMapping):
	__slots__ = ("graph",)

	class Successors(AbstractSuccessors):
		__slots__ = ("graph", "container", "orig", "_cache")

		def _order_nodes(self, dest):
			return (self.orig, dest)


class DiGraphPredecessorsMapping(GraphEdgeMapping):
	"""Mapping for Predecessors instances, which map to Edges that end at
	the dest provided to this

	"""

	__slots__ = ("graph",)

	def __contains__(self, dest):
		for orig in self.db._edges_cache.iter_predecessors(
			self.graph.name, dest, *self.db._btt()
		):
			try:
				if self.db._edges_cache.retrieve(
					self.graph.name, orig, dest, *self.db._btt()
				):
					return True
			except KeyError:
				continue
		return False

	def __getitem__(self, dest):
		"""Return a Predecessors instance for edges ending at the given
		node

		"""
		if dest not in self.graph.node:
			raise KeyError("No such node", dest)
		if dest not in self._cache:
			self._cache[dest] = self.Predecessors(self, dest)
		return self._cache[dest]

	def __setitem__(self, key, val):
		"""Interpret ``val`` as a mapping of edges that end at ``dest``"""
		created = key not in self
		if key not in self._cache:
			self._cache[key] = self.Predecessors(self, key)
		preds = self._cache[key]
		preds.clear()
		preds.update(val)

	def __delitem__(self, key):
		"""Delete all edges ending at ``dest``"""
		it = self[key]
		it.clear()
		del self._cache[key]

	def __iter__(self):
		return iter(self.graph.node)

	def __len__(self):
		return len(self.graph.node)

	class Predecessors(GraphEdgeMapping):
		"""Mapping of Edges that end at a particular node"""

		__slots__ = ("graph", "container", "dest")

		def __init__(self, container, dest):
			"""Store container and node ID"""
			super().__init__(container.graph)
			self.container = container
			self.dest = dest

		def __iter__(self):
			"""Iterate over the edges that exist at the present (branch, rev)"""
			for orig in self.db._edges_cache.iter_predecessors(
				self.graph.name, self.dest, *self.db._btt()
			):
				if orig in self:
					yield orig

		def __contains__(self, orig):
			"""Is there an edge from ``orig`` at the moment?"""
			return self.db._edges_cache.has_predecessor(
				self.graph.name, self.dest, orig, *self.db._btt()
			)

		def __len__(self):
			"""How many edges exist at this rev of this branch?"""
			n = 0
			for n, _ in enumerate(self, start=1):
				pass
			return n

		def _make_edge(self, orig):
			return Edge(self.graph, orig, self.dest)

		def __getitem__(self, orig):
			"""Get the edge from the given node to mine"""
			if orig not in self:
				raise KeyError(orig)
			return self.graph.adj[orig][self.dest]

		def __setitem__(self, orig, value):
			"""Use ``value`` as a mapping of edge attributes, set an edge from the
			given node to mine.

			"""
			branch, turn, tick = self.db._nbtt()
			try:
				e = self[orig]
				e.clear()
			except KeyError:
				self.db.query.exist_edge(
					self.graph.name,
					orig,
					self.dest,
					branch,
					turn,
					tick,
					True,
				)
				e = self._make_edge(orig)
			e.update(value)
			self.db._edges_cache.store(
				self.graph.name, orig, self.dest, branch, turn, tick, True
			)

		def __delitem__(self, orig):
			"""Unset the existence of the edge from the given node to mine"""
			branch, turn, tick = self.db._nbtt()
			self.db.query.exist_edge(
				self.graph.name, orig, self.dest, branch, turn, tick, False
			)
			self.db._edges_cache.store(
				self.graph.name, orig, self.dest, branch, turn, tick, None
			)


def unwrapped_dict(d):
	ret = {}
	for k, v in d.items():
		if hasattr(v, "unwrap") and not getattr(v, "no_unwrap", False):
			ret[k] = v.unwrap()
		else:
			ret[k] = v
	return ret


class DiGraph(networkx.DiGraph, ABC):
	"""A version of the networkx.DiGraph class that stores its state in a
	database.

	"""

	adj_cls = DiGraphSuccessorsMapping
	pred_cls = DiGraphPredecessorsMapping
	graph_map_cls = GraphMapping
	node_map_cls = GraphNodeMapping
	_statmap: graph_map_cls
	_nodemap: node_map_cls
	_adjmap: adj_cls
	_predmap: pred_cls

	def __repr__(self):
		return "<{} object named {} containing {} nodes, {} edges>".format(
			self.__class__, self.name, len(self.nodes), len(self.edges)
		)

	def _nodes_state(self):
		return {
			noden: {
				k: v for (k, v) in unwrapped_dict(node).items() if k != "name"
			}
			for noden, node in self._node.items()
		}

	def _edges_state(self):
		ret = {}
		ismul = self.is_multigraph()
		for orig, dests in self.adj.items():
			if orig not in ret:
				ret[orig] = {}
			origd = ret[orig]
			for dest, edge in dests.items():
				if ismul:
					if dest not in origd:
						origd[dest] = edges = {}
					else:
						edges = origd[dest]
					for i, val in edge.items():
						edges[i] = unwrapped_dict(val)
				else:
					origd[dest] = unwrapped_dict(edge)
		return ret

	def _val_state(self):
		return {
			k: v
			for (k, v) in unwrapped_dict(self.graph).items()
			if k != "name"
		}

	def __init__(self, db, name):  # user shouldn't instantiate directly
		self._name = name
		self.db = db

	def __bool__(self):
		return self._name in self.db._graph_objs

	@property
	def graph(self):
		if not hasattr(self, "_statmap"):
			self._statmap = self.graph_map_cls(self)
		return self._statmap

	@graph.setter
	def graph(self, v):
		self.graph.clear()
		self.graph.update(v)

	@property
	def node(self):
		if not hasattr(self, "_nodemap"):
			self._nodemap = self.node_map_cls(self)
		return self._nodemap

	_node = node

	@property
	def adj(self):
		if not hasattr(self, "_adjmap"):
			self._adjmap = self.adj_cls(self)
		return self._adjmap

	edge = succ = _succ = _adj = adj

	@property
	def pred(self):
		if not hasattr(self, "pred_cls"):
			raise TypeError("Undirected graph")
		if not hasattr(self, "_predmap"):
			self._predmap = self.pred_cls(self)
		return self._predmap

	_pred = pred

	@property
	def name(self):
		return self._name

	@name.setter
	def name(self, v):
		raise TypeError("graphs can't be renamed")

	def remove_node(self, n):
		"""Version of remove_node that minimizes writes"""
		if n not in self._node:
			raise NetworkXError("The node %s is not in the digraph." % (n,))
		nbrs = list(self._succ[n])
		for u in nbrs:
			del self._pred[u][n]  # remove all edges n-u in digraph
		pred = list(self._pred[n])
		for u in pred:
			del self._succ[u][n]  # remove all edges n-u in digraph
		del self._node[n]

	def remove_edge(self, u, v):
		"""Version of remove_edge that's much like normal networkx but only
		deletes once, since the database doesn't keep separate adj and
		succ mappings

		"""
		try:
			del self.succ[u][v]
		except KeyError:
			raise NetworkXError(
				"The edge {}-{} is not in the graph.".format(u, v)
			)

	def remove_edges_from(self, ebunch):
		"""Version of remove_edges_from that's much like normal networkx but only
		deletes once, since the database doesn't keep separate adj and
		succ mappings

		"""
		for e in ebunch:
			(u, v) = e[:2]
			if u in self.succ and v in self.succ[u]:
				del self.succ[u][v]

	def add_edge(self, u, v, attr_dict=None, **attr):
		"""Version of add_edge that only writes to the database once"""
		if attr_dict is None:
			attr_dict = attr
		else:
			try:
				attr_dict.update(attr)
			except AttributeError:
				raise NetworkXError(
					"The attr_dict argument must be a dictionary."
				)
		if u not in self.node:
			self.node[u] = {}
		if v not in self.node:
			self.node[v] = {}
		if u in self.adj:
			datadict = self.adj[u].get(v, {})
		else:
			self.adj[u] = {v: {}}
			datadict = self.adj[u][v]
		datadict.update(attr_dict)
		self.succ[u][v] = datadict

	def add_edges_from(self, ebunch, attr_dict=None, **attr):
		"""Version of add_edges_from that only writes to the database once"""
		if attr_dict is None:
			attr_dict = attr
		else:
			try:
				attr_dict.update(attr)
			except AttributeError:
				raise NetworkXError("The attr_dict argument must be a dict.")
		for e in ebunch:
			ne = len(e)
			if ne == 3:
				u, v, dd = e
				assert hasattr(dd, "update")
			elif ne == 2:
				u, v = e
				dd = {}
			else:
				raise NetworkXError(
					"Edge tupse {} must be a 2-tuple or 3-tuple.".format(e)
				)
			if u not in self.node:
				self.node[u] = {}
			if v not in self.node:
				self.node[v] = {}
			datadict = self.adj.get(u, {}).get(v, {})
			datadict.update(attr_dict)
			datadict.update(dd)
			self.succ[u][v] = datadict
			assert u in self.succ
			assert v in self.succ[u]

	def clear(self):
		"""Remove all nodes and edges from the graph.

		Unlike the regular networkx implementation, this does *not*
		remove the graph's name. But all the other graph, node, and
		edge attributes go away.

		"""
		self.adj.clear()
		self.node.clear()
		self.graph.clear()

	def add_node(self, node_for_adding, **attr):
		"""Version of add_node that minimizes writes"""
		if node_for_adding not in self._succ:
			self._succ[node_for_adding] = self.adjlist_inner_dict_factory()
			self._pred[node_for_adding] = self.adjlist_inner_dict_factory()
			self._node[node_for_adding] = self.node_dict_factory()
		self._node[node_for_adding].update(attr)


class GraphsMapping(MutableMapping):
	def __init__(self, orm):
		self.orm = orm

	def __iter__(self):
		"""Iterate over every character name."""
		for name in self.orm._graph_objs:
			if name in self:
				yield name

	def __contains__(self, name):
		"""Has this character been created?"""
		try:
			return (
				self.orm._graph_cache.retrieve(name, *self.orm._btt())
				!= "Deleted"
			)
		except KeyError:
			return False

	def __len__(self):
		"""How many characters have been created?"""
		n = 0
		for name in self.orm._graph_objs:
			if name in self:
				n += 1
		return n

	def __getitem__(self, item):
		if not self.orm._has_graph(item):
			raise KeyError(f"No such graph: {item}", item)
		return self.orm._graph_objs[item]

	def __setitem__(self, key, value):
		self.orm.new_character(key, data=value)

	def __delitem__(self, key):
		if key not in self:
			raise KeyError("No such graph")
		self.orm.query.del_graph(key, *self.orm._btt())
		if key in self.orm._graph_objs:
			del self.orm._graph_objs[key]



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.f
from __future__ import annotations

from collections.abc import Iterable
from types import GenericAlias
from typing import (
	Annotated,
	Any,
	Callable,
	Literal,
	NewType,
	TypeAlias,
	TypeGuard,
)

import networkx as nx
from annotated_types import Ge

from .wrap import DictWrapper, ListWrapper, SetWrapper

_Key = str | int | float | None | tuple["_Key", ...] | frozenset["_Key"]


def is_valid_key(obj: _Key) -> TypeGuard[Key]:
	"""Is this an object that Lisien can serialize as a key?"""
	return (
		obj is None
		or isinstance(obj, (str, int, float))
		or (
			isinstance(obj, (tuple, frozenset))
			and all(is_valid_key(elem) for elem in obj)
		)
	)


class _KeyMeta(type):
	def __instancecheck__(self, instance) -> TypeGuard[Key]:
		return is_valid_key(instance)

	def __call__(self, obj: _Key) -> Key:
		if is_valid_key(obj):
			return obj
		raise TypeError("Not a valid key", obj)

	def __class_getitem__(cls, item):
		return GenericAlias(cls, item)


class Key(metaclass=_KeyMeta):
	def __new__(cls, obj: _Key) -> Key:
		if not is_valid_key(obj):
			raise TypeError("Invalid key")
		return obj


_Value: TypeAlias = (
	_Key
	| dict[_Key, "_Value"]
	| tuple["_Value", ...]
	| list["_Value"]
	| set["_Value"]
	| frozenset["_Value"]
	| DictWrapper
	| ListWrapper
	| SetWrapper
	| type(...)
)


def is_valid_value(obj: _Value) -> TypeGuard[Value]:
	"""Is this an object that Lisien can serialize as a value?"""
	return (
		obj is ...
		or is_valid_key(obj)
		or (
			isinstance(obj, (dict, DictWrapper))
			and all(map(is_valid_key, obj.keys()))
			and all(map(is_valid_value, obj.values()))
		)
		or (
			isinstance(
				obj, (tuple, list, set, frozenset, ListWrapper, SetWrapper)
			)
			and isinstance(obj, Iterable)
			and all(map(is_valid_value, obj))
		)
		or (
			isinstance(obj, nx.DiGraph)
			and all(map(is_valid_key, obj.graph.keys()))
			and all(map(is_valid_value, obj.graph.values()))
			and all(
				is_valid_key(k) and is_valid_value(v)
				for node in obj.nodes().values()
				for (k, v) in node.items()
			)
			and all(
				is_valid_key(orig)
				and is_valid_key(dest)
				and is_valid_key(k)
				and is_valid_value(v)
				for orig in obj.adj
				for dest in obj.adj[orig]
				for (k, v, *_) in obj.adj[orig][dest]
			)
		)
	)


class _ValueMeta(type):
	def __instancecheck__(self, instance) -> TypeGuard[Value]:
		return is_valid_value(instance)

	def __call__(self, obj: _Value) -> Value:
		if is_valid_value(obj):
			return obj
		raise TypeError("Not a valid value", obj)

	def __class_getitem__(cls, item):
		return GenericAlias(cls, item)


class Value(metaclass=_ValueMeta):
	def __new__(cls, obj: _Value) -> Value:
		if not is_valid_value(obj):
			raise TypeError("Invalid value")
		return obj


Stat = NewType("Stat", Key)
EternalKey = NewType("EternalKey", Key)
UniversalKey = NewType("UniversalKey", Key)
Branch = NewType("Branch", str)
Turn = NewType("Turn", Annotated[int, Ge(0)])
Tick = NewType("Tick", Annotated[int, Ge(0)])
Time: TypeAlias = tuple[Branch, Turn, Tick]
LinearTime: TypeAlias = tuple[Turn, Tick]
TimeWindow: TypeAlias = tuple[Branch, Turn, Tick, Turn, Tick]
Plan = NewType("Plan", Annotated[int, Ge(0)])
CharName = NewType("CharName", Key)
NodeName = NewType("NodeName", Key)
EntityKey: TypeAlias = (
	tuple[CharName]
	| tuple[CharName, NodeName]
	| tuple[CharName, NodeName, NodeName]
)
RulebookName = NewType("RulebookName", Key)
RulebookPriority = NewType("RulebookPriority", float)
RuleName = NewType("RuleName", str)
RuleNeighborhood: TypeAlias = Annotated[int, Ge(0)] | None
RuleBig = NewType("RuleBig", bool)
RuleFunc: TypeAlias = Callable[[Any], bool]
FuncName = NewType("FuncName", str)
FuncStoreName: TypeAlias = Literal[
	"trigger", "prereq", "action", "function", "method"
]
TriggerFuncName = NewType("TriggerFuncName", FuncName)
PrereqFuncName = NewType("PrereqFuncName", FuncName)
ActionFuncName = NewType("ActionFuncName", FuncName)
RuleFuncName: TypeAlias = TriggerFuncName | PrereqFuncName | ActionFuncName
UniversalKeyframe = NewType("UniversalKeyframe", dict)
RuleKeyframe = NewType("RuleKeyframe", dict)
RulebookKeyframe = NewType("RulebookKeyframe", dict)
NodeKeyframe = NewType("NodeKeyframe", dict)
EdgeKeyframe = NewType("EdgeKeyframe", dict)
NodeRowType: TypeAlias = tuple[CharName, NodeName, Branch, Turn, Tick, bool]
EdgeRowType: TypeAlias = tuple[
	CharName, NodeName, NodeName, Branch, Turn, Tick, bool
]
GraphValRowType: TypeAlias = tuple[CharName, Key, Branch, Turn, Tick, Value]
NodeValRowType: TypeAlias = tuple[
	CharName, NodeName, Key, Branch, Turn, Tick, Value
]
EdgeValRowType: TypeAlias = tuple[
	CharName, NodeName, NodeName, Key, Branch, Turn, Tick, Value
]
StatDict: TypeAlias = dict[Stat | Literal["rulebook"], Value]
CharDict: TypeAlias = dict[
	Stat
	| Literal[
		"units",
		"character_rulebook",
		"unit_rulebook",
		"character_thing_rulebook",
		"character_place_rulebook",
		"character_portal_rulebook",
	],
	Value,
]
GraphValKeyframe: TypeAlias = dict[CharName, CharDict]
NodeValDict: TypeAlias = dict[NodeName, StatDict]
GraphNodeValKeyframe: TypeAlias = dict[CharName, NodeValDict]
EdgeValDict: TypeAlias = dict[NodeName, dict[NodeName, StatDict]]
GraphEdgeValKeyframe: TypeAlias = dict[CharName, EdgeValDict]
NodesDict: TypeAlias = dict[NodeName, bool]
GraphNodesKeyframe: TypeAlias = dict[CharName, NodesDict]
EdgesDict: TypeAlias = dict[NodeName, dict[NodeName, bool]]
GraphEdgesKeyframe: TypeAlias = dict[CharName, EdgesDict]
UnitsDict: TypeAlias = dict[CharName, dict[NodeName, bool]]
CharDelta: TypeAlias = dict[
	Stat
	| Literal[
		"character_rulebook",
		"unit_rulebook",
		"character_thing_rulebook",
		"character_place_rulebook",
		"character_portal_rulebook",
		"nodes",
		"node_val",
		"edges",
		"edge_val",
		"rulebooks",
		"units",
	],
	NodesDict
	| NodeValDict
	| EdgesDict
	| EdgeValDict
	| RulebookName
	| UnitsDict
	| Value,
]
DeltaDict: TypeAlias = dict[
	CharName,
	CharDelta | None,
]
KeyframeTuple: TypeAlias = tuple[
	CharName,
	Branch,
	Turn,
	Tick,
	GraphNodeValKeyframe,
	GraphEdgeValKeyframe,
	GraphValKeyframe,
]
Keyframe: TypeAlias = dict[
	Literal[
		"universal",
		"triggers",
		"prereqs",
		"actions",
		"neighborhood",
		"big",
		"rulebook",
		"nodes",
		"edges",
		"node_val",
		"edge_val",
		"graph_val",
	],
	GraphValKeyframe
	| GraphNodesKeyframe
	| GraphNodeValKeyframe
	| GraphEdgesKeyframe
	| GraphEdgeValKeyframe
	| dict[UniversalKey, Value]
	| dict[RuleName, list[TriggerFuncName]]
	| dict[RuleName, list[PrereqFuncName]]
	| dict[RuleName, list[ActionFuncName]]
	| dict[RuleName, int]
	| dict[RuleName, bool]
	| dict[RulebookName, RulebookKeyframe],
]
SlightlyPackedDeltaType: TypeAlias = dict[
	bytes,
	dict[
		bytes,
		bytes
		| dict[
			bytes,
			bytes | dict[bytes, bytes | dict[bytes, bytes]],
		],
	],
]
RulebookTypeStr: TypeAlias = Literal[
	"character",
	"unit",
	"character_thing",
	"character_place",
	"character_portal",
]
CharacterRulebookTypeStr: TypeAlias = Literal[
	"character_rulebook",
	"unit_rulebook",
	"character_thing_rulebook",
	"character_place_rulebook",
	"character_portal_rulebook",
]



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" The fundamental unit of game logic, the Rule, and structures to
store and organize them in.

A Rule is three lists of functions: triggers, prereqs, and actions.
The actions do something, anything that you need your game to do, but
probably making a specific change to the world model. The triggers and
prereqs between them specify when the action should occur: any of its
triggers can tell it to happen, but then any of its prereqs may stop it
from happening.

Rules are assembled into RuleBooks, essentially just lists of Rules
that can then be assigned to be followed by any game entity --
but each game entity has its own RuleBook by default, and you never
need to change that.

To add a new rule to a lisien entity, the easiest thing is to use the
decorator syntax::

	@entity.rule
	def do_something(entity):
		...

	@do_something.trigger
	def whenever(entity):
		...

	@do_something.trigger
	def forever(entity):
		...

	@do_something.action
	def do_something_else(entity):
		...



When run, this code will:

	* copy the ``do_something`` function to ``action.py``, where lisien knows\
			to run it when a rule triggers it

	* create a new rule named ``'do_something'``

	* set the function ``do_something`` as the first (and, so far, only) entry\
			in the actions list of the rule by that name

	* copy the ``whenever`` function to ``trigger.py``, where lisien knows to\
			call it when a rule has it as a trigger

	* set the function ``whenever`` as the first entry in the triggers list\
			of the rule ``'do_something'``

	* append the function ``forever`` to the same triggers list

	* copy ``do_something_else`` to ``action.py``

	* append ``do_something_else`` to the actions list of the rule

The ``trigger``, ``prereq``, and ``action`` attributes of Rule objects
may also be used like lists. You can put functions in them yourself,
provided they are already present in the correct module. If it's
inconvenient to get the actual function object, use a string of
the function's name.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from ast import parse
from collections.abc import Hashable, Iterable, MutableMapping, MutableSequence
from functools import cached_property, partial
from typing import TYPE_CHECKING, Any, Callable, Iterator, Literal, Optional

from astunparse import unparse
from blinker import Signal

from .cache import FuncListCache
from .types import (
	ActionFuncName,
	Branch,
	Key,
	PrereqFuncName,
	RuleBig,
	RulebookName,
	RulebookPriority,
	RuleFunc,
	RuleFuncName,
	RuleName,
	RuleNeighborhood,
	Tick,
	TriggerFuncName,
	Turn,
	Value,
)
from .util import AbstractEngine, dedent_source
from .xcollections import FunctionStore

if TYPE_CHECKING:
	from .engine import Engine


def roundtrip_dedent(source):
	"""Reformat some lines of code into what unparse makes."""
	return unparse(parse(dedent_source(source)))


class RuleFuncList(MutableSequence, Signal, ABC):
	"""Abstract class for lists of functions like trigger, prereq, action"""

	__slots__ = ["rule"]
	_funcstore: FunctionStore
	_cache: FuncListCache
	_setter: Callable[[RuleName, Branch, Turn, Tick, list[RuleFuncName]], None]

	def __init__(self, rule: Rule):
		super().__init__()
		self.rule = rule

	def __repr__(self):
		return f"<class 'lisien.rule.{self.__class__.__name__}' [{', '.join(self._get())}]>"

	def _nominate(self, v: RuleFunc | str | RuleFuncName) -> RuleFuncName:
		if callable(v):
			self._funcstore(v)
			return v.__name__
		if not hasattr(self._funcstore, v):
			raise KeyError(
				"No function by that name in this store", v, self._funcstore
			)
		return v

	def _get(self) -> list[RuleFuncName]:
		return self._cache.retrieve(self.rule.name, *self.rule.engine._btt())

	def _set(self, v: list[RuleFuncName]) -> None:
		branch, turn, tick = self.rule.engine._nbtt()
		self._cache.store(Key(self.rule.name), branch, turn, tick, v)
		self._setter(self.rule.name, branch, turn, tick, v)

	def __iter__(self):
		for funcname in self._get():
			yield getattr(self._funcstore, funcname)

	def __contains__(self, item: RuleFunc | str | RuleFuncName):
		if hasattr(item, "__name__"):
			item = item.__name__
		return item in self._get()

	def __len__(self):
		return len(self._get())

	def __getitem__(self, i: int):
		return getattr(self._funcstore, self._get()[i])

	def __setitem__(self, i: int, v: RuleFunc | str | RuleFuncName):
		v = self._nominate(v)
		l = list(self._get())
		l[i] = v
		self._set(list(l))
		self.send(self)

	def __delitem__(self, i: int):
		l = list(self._get())
		del l[i]
		self._set(list(l))
		self.send(self)

	def insert(self, i: int, v: RuleFunc | str | RuleFuncName) -> None:
		l = list(self._get())
		l.insert(i, self._nominate(v))
		self._set(list(l))
		self.send(self)

	def append(self, v: RuleFunc | str | RuleFuncName) -> None:
		try:
			old = self._get()
		except KeyError:
			old = []
		self._set(old + [self._nominate(v)])
		self.send(self)

	def index(
		self,
		x: RuleFunc | str | RuleFuncName,
		start: int = 0,
		end: int | None = None,
	):
		if not callable(x):
			x = getattr(self._funcstore, x)
		return super().index(x, start, end)


class TriggerList(RuleFuncList):
	"""A list of trigger functions for rules"""

	@cached_property
	def _funcstore(self) -> FunctionStore:
		return self.rule.engine.trigger

	@cached_property
	def _cache(self) -> InitializedEntitylessCache:
		return self.rule.engine._triggers_cache

	@cached_property
	def _setter(
		self,
	) -> Callable[[RuleName, Branch, Turn, Tick, list[TriggerFuncName]], None]:
		return self.rule.engine.query.set_rule_triggers


class PrereqList(RuleFuncList):
	"""A list of prereq functions for rules"""

	@cached_property
	def _funcstore(self) -> FunctionStore:
		return self.rule.engine.prereq

	@cached_property
	def _cache(self) -> InitializedEntitylessCache:
		return self.rule.engine._prereqs_cache

	@cached_property
	def _setter(
		self,
	) -> Callable[[RuleName, Branch, Turn, Tick, list[PrereqFuncName]], None]:
		return self.rule.engine.query.set_rule_prereqs


class ActionList(RuleFuncList):
	"""A list of action functions for rules"""

	@cached_property
	def _funcstore(self) -> FunctionStore:
		return self.rule.engine.action

	@cached_property
	def _cache(self) -> InitializedEntitylessCache:
		return self.rule.engine._actions_cache

	@cached_property
	def _setter(
		self,
	) -> Callable[[RuleName, Branch, Turn, Tick, list[ActionFuncName]], None]:
		return self.rule.engine.query.set_rule_actions


class RuleFuncListDescriptor:
	"""Descriptor that lets you get and set a whole RuleFuncList at once"""

	__slots__ = ("cls",)

	def __init__(self, cls):
		self.cls = cls

	@property
	def flid(self):
		return "_funclist" + str(id(self))

	def __get__(self, obj: Rule, type=None) -> RuleFuncList:
		if not hasattr(obj, self.flid):
			setattr(obj, self.flid, self.cls(obj))
		return getattr(obj, self.flid)

	def __set__(self, obj: Rule, value: list[RuleFunc | str | RuleFuncName]):
		if not hasattr(obj, self.flid):
			setattr(obj, self.flid, self.cls(obj))
		flist = getattr(obj, self.flid)
		namey_value = [flist._nominate(v) for v in value]
		flist._set(namey_value)
		flist.send(flist)

	def __delete__(self, obj):
		raise TypeError("Rules must have their function lists")


class Rule:
	"""Stuff that might happen in the simulation under some conditions

	Rules are comprised of three lists of functions:

	* actions, which mutate the world state
	* triggers, which make the actions happen
	* prereqs, which prevent the actions from happening when triggered

	Each kind of function should be stored in the appropriate module
	supplied to the lisien core at startup. This makes it possible to
	load the functions on later startups. You may instead use the string
	name of a function already stored in the module, or use the
	``trigger``, ``prereq``, or ``action`` decorator on a new function to
	add it to both the module and the rule.

	"""

	triggers: TriggerList = RuleFuncListDescriptor(TriggerList)
	prereqs: PrereqList = RuleFuncListDescriptor(PrereqList)
	actions: ActionList = RuleFuncListDescriptor(ActionList)
	name: RuleName

	@property
	def neighborhood(self) -> RuleNeighborhood:
		try:
			ret = self.engine._neighborhoods_cache.retrieve(
				self.name, *self.engine._btt()
			)
		except KeyError:
			return None
		return ret

	@neighborhood.setter
	def neighborhood(self, neighbors: RuleNeighborhood):
		if neighbors is not None and not isinstance(neighbors, int):
			raise TypeError("Not an int", neighbors)
		if neighbors < 0:
			raise ValueError("Can't be negative", neighbors)
		btt = self.engine._nbtt()
		self.engine._neighborhoods_cache.store(
			Key(self.name), *btt, Value(neighbors)
		)
		self.engine.query.set_rule_neighborhood(self.name, *btt, neighbors)

	@property
	def big(self) -> RuleBig:
		try:
			return self.engine._rule_bigness_cache.retrieve(
				self.name, *self.engine._btt()
			)
		except KeyError:
			return RuleBig(False)

	@big.setter
	def big(self, big: bool | RuleBig):
		btt = self.engine._nbtt()
		self.engine._rule_bigness_cache.store(Key(self.name), *btt, Value(big))
		self.engine.query.set_rule_big(self.name, *btt, big)

	def __init__(
		self,
		engine: "Engine",
		name: RuleName | str,
		triggers: list[TriggerFuncName | Callable[[Any], bool] | str]
		| None = None,
		prereqs: list[PrereqFuncName | Callable[[Any], bool] | str]
		| None = None,
		actions: list[ActionFuncName | Callable[[Any], bool] | str]
		| None = None,
		neighborhood: RuleNeighborhood = None,
		big: RuleBig = RuleBig(False),
		create: bool = True,
	):
		"""Store the engine and my name, make myself a record in the database
		if needed, and instantiate one FunList each for my triggers,
		actions, and prereqs.

		"""
		self.engine = engine
		self.name = self.__name__ = RuleName(name)
		# It'd be more correct to subclass InitializedEntitylessCache and
		# give all its methods more specific signatures that take RuleName.
		name = Key(name)
		if create:
			branch, turn, tick = engine._nbtt()
			if (
				self.engine._triggers_cache.contains_key(
					name, branch, turn, tick
				)
				or self.engine._prereqs_cache.contains_key(
					name, branch, turn, tick
				)
				or self.engine._actions_cache.contains_key(
					name, branch, turn, tick
				)
			):
				(branch, turn, tick) = self.engine._nbtt()
			triggers = list(self._fun_names_iter("trigger", triggers or []))
			prereqs = list(self._fun_names_iter("prereq", prereqs or []))
			actions = list(self._fun_names_iter("action", actions or []))
			self.engine.query.set_rule(
				name,
				branch,
				turn,
				tick,
				triggers,
				prereqs,
				actions,
				neighborhood,
				big,
			)
			self.engine._triggers_cache.store(
				name, branch, turn, tick, triggers
			)
			self.engine._prereqs_cache.store(name, branch, turn, tick, prereqs)
			self.engine._actions_cache.store(name, branch, turn, tick, actions)
			self.engine._neighborhoods_cache.store(
				name, branch, turn, tick, neighborhood
			)
			self.engine._rule_bigness_cache.store(
				name, branch, turn, tick, big
			)
			# Don't *make* a keyframe -- but if there happens to already *be*
			# a keyframe at this very moment, add the new rule to it
			if (branch, turn, tick) in self.engine._keyframes_times:
				# ensure it's loaded
				self.engine._get_keyframe(branch, turn, tick, silent=True)
				# Just because there's a keyframe doesn't mean it's in every cache.
				# I should probably change that.
				try:
					trigkf = self.engine._triggers_cache.get_keyframe(
						branch, turn, tick
					)
				except KeyError:
					trigkf = {
						aname: self.engine._triggers_cache.retrieve(
							aname, branch, turn, tick
						)
						for aname in self.engine._triggers_cache.iter_keys(
							branch, turn, tick
						)
					}
				try:
					preqkf = self.engine._prereqs_cache.get_keyframe(
						branch, turn, tick
					)
				except KeyError:
					preqkf = {
						aname: self.engine._prereqs_cache.retrieve(
							aname, branch, turn, tick
						)
						for aname in self.engine._prereqs_cache.iter_keys(
							branch, turn, tick
						)
					}
				try:
					actkf = self.engine._actions_cache.get_keyframe(
						branch, turn, tick
					)
				except KeyError:
					actkf = {
						aname: self.engine._actions_cache.retrieve(
							aname, branch, turn, tick
						)
						for aname in self.engine._actions_cache.iter_keys(
							branch, turn, tick
						)
					}
				try:
					nbrkf = self.engine._neighborhoods_cache.get_keyframe(
						branch, turn, tick
					)
				except KeyError:
					nbrkf = {
						aname: self.engine._neighborhoods_cache.retrieve(
							aname, branch, turn, tick
						)
						for aname in self.engine._neighborhoods_cache.iter_keys(
							branch, turn, tick
						)
					}
				try:
					bigkf = self.engine._rule_bigness_cache.get_keyframe(
						branch, turn, tick
					)

				except KeyError:
					bigkf = {
						aname: self.engine._rule_bigness_cache.retrieve(
							aname, branch, turn, tick
						)
						for aname in self.engine._rule_bigness_cache.iter_keys(
							branch, turn, tick
						)
					}
				trigkf[name] = triggers
				preqkf[name] = prereqs
				actkf[name] = actions
				nbrkf[name] = neighborhood
				bigkf[name] = big
				self.engine._triggers_cache.set_keyframe(
					branch, turn, tick, trigkf
				)
				self.engine._prereqs_cache.set_keyframe(
					branch, turn, tick, preqkf
				)
				self.engine._actions_cache.set_keyframe(
					branch, turn, tick, actkf
				)
				self.engine._neighborhoods_cache.set_keyframe(
					branch, turn, tick, nbrkf
				)
				self.engine._rule_bigness_cache.set_keyframe(
					branch, turn, tick, bigkf
				)

	def __eq__(self, other: Rule):
		return isinstance(other, Rule) and self.name == other.name

	def _fun_names_iter(
		self,
		functyp: Literal["trigger", "prereq", "action"],
		val: Iterable[RuleFunc | str | RuleFuncName],
	):
		"""Iterate over the names of the functions in ``val``,
		adding them to ``funcstore`` if they are missing;
		or if the items in ``val`` are already the names of functions
		in ``funcstore``, iterate over those.

		"""
		funcstore = getattr(self.engine, functyp)
		for v in val:
			if callable(v):
				if v != getattr(funcstore, v.__name__, None):
					setattr(funcstore, v.__name__, v)
				yield v.__name__
			elif v not in funcstore:
				raise KeyError(
					"Function {} not present in {}".format(v, funcstore._tab)
				)
			else:
				yield v

	def __repr__(self):
		return f"<Rule({repr(self.name)})>"

	def trigger(
		self, fun: RuleFunc | str | TriggerFuncName
	) -> RuleFunc | str | TriggerFuncName:
		"""Decorator to append the function to my triggers list."""
		self.triggers.append(fun)
		return fun

	def prereq(
		self, fun: RuleFunc | str | PrereqFuncName
	) -> RuleFunc | str | PrereqFuncName:
		"""Decorator to append the function to my prereqs list."""
		self.prereqs.append(fun)
		return fun

	def action(
		self, fun: RuleFunc | str | ActionFuncName
	) -> RuleFunc | str | ActionFuncName:
		"""Decorator to append the function to my actions list."""
		self.actions.append(fun)
		return fun

	def always(self) -> None:
		"""Arrange to be triggered every turn"""
		self.triggers = [self.engine.trigger.truth]


class RuleBook(MutableSequence, Signal):
	"""A list of rules to be followed for some Character, or a part of it"""

	def _get_cache(
		self, branch: Branch, turn: Turn, tick: Tick
	) -> tuple[list[RuleName], RulebookPriority]:
		rules: list[RuleName] = []
		prio = RulebookPriority(0.0)
		try:
			rules, prio = self.engine._rulebooks_cache.retrieve(
				self.name, branch, turn, tick
			)
		except KeyError:
			pass
		return list(rules), prio

	def _set_cache(
		self,
		branch: Branch,
		turn: Turn,
		tick: Tick,
		v: tuple[list[RuleName], RulebookPriority],
	) -> None:
		with self.engine._rulebooks_cache.overwriting():
			self.engine._rulebooks_cache.store(
				self.name, branch, turn, tick, v
			)

	def __init__(self, engine: "Engine", name: RulebookName):
		super().__init__()
		self.engine = engine
		self.name = name

	@property
	def priority(self) -> RulebookPriority:
		return self._get_cache(*self.engine._btt())[1]

	@priority.setter
	def priority(self, v: float | RulebookPriority):
		v = RulebookPriority(float(v))
		branch, turn, tick = self.engine._nbtt()
		cache, _ = self._get_cache(branch, turn, tick)
		self._set_cache(branch, turn, tick, (cache, v))
		self.engine.query.set_rulebook(self.name, branch, turn, tick, cache, v)

	def __contains__(self, v: RuleName | Rule) -> bool:
		return getattr(v, "name", v) in self._get_cache(*self.engine._btt())[0]

	def __iter__(self) -> Iterator[Rule]:
		rule_name: RuleName
		for rule_name in self._get_cache(*self.engine._btt())[0]:
			yield self.engine.rule[rule_name]

	def __len__(self):
		try:
			return len(self._get_cache(*self.engine._btt())[0])
		except KeyError:
			return 0

	def __getitem__(self, i: int) -> Rule:
		return self.engine.rule[self._get_cache(*self.engine._btt())[0][i]]

	def _coerce_rule(self, v: Rule | str | RuleName) -> Rule:
		if isinstance(v, Rule):
			return v
		elif isinstance(v, str):
			return self.engine.rule[v]
		else:
			raise TypeError("Invalid rule name", v)

	def __setitem__(self, i: int, v: Rule | str | RuleName):
		v = RuleName(getattr(v, "name", v))
		if v == "truth":
			raise ValueError("Illegal rule name")
		branch, turn, tick = self.engine._nbtt()
		cache, prio = self._get_cache(branch, turn, tick)
		try:
			cache[i] = v
		except IndexError:
			if i != len(cache):
				raise
			if i == 0:
				cache = [v]
			else:
				assert i == len(cache)
				cache.append(v)
			prio = RulebookPriority(0.0)
			self._set_cache(branch, turn, tick, (cache, prio))
		self.engine.query.set_rulebook(
			self.name, branch, turn, tick, cache, prio
		)
		self.engine.rulebook.send(self, i=i, v=v)
		self.send(self, i=i, v=v)

	def insert(self, i: int, v: Rule | str | RuleName) -> None:
		v = RuleName(getattr(v, "name", v))
		if v == "truth":
			raise ValueError("Illegal rule name")
		branch, turn, tick = self.engine._nbtt()
		try:
			cache, prio = self._get_cache(branch, turn, tick)
			cache.insert(i, v)
		except KeyError:
			if i != 0:
				raise IndexError
			cache = [v]
			prio = RulebookPriority(0.0)
		self._set_cache(branch, turn, tick, (cache, prio))
		self.engine.query.set_rulebook(
			self.name, branch, turn, tick, cache, prio
		)
		self.engine.rulebook.send(self, i=i, v=v)
		self.send(self, i=i, v=v)

	def index(
		self, v: Rule | str | RuleName, start: int = 0, stop: int | None = None
	) -> int:
		args = [v, start]
		if stop is not None:
			args.append(stop)
		if isinstance(v, str):
			try:
				return self._get_cache(*self.engine._btt())[0].index(*args)
			except KeyError:
				raise ValueError
		return super().index(*args)

	def __delitem__(self, i: int) -> None:
		branch, turn, tick = self.engine._btt()
		try:
			cache, prio = self._get_cache(branch, turn, tick)
		except KeyError:
			raise IndexError
		del cache[i]
		self.engine.query.set_rulebook(
			self.name, branch, turn, tick, cache, prio
		)
		self._set_cache(branch, turn, tick, (cache, prio))
		self.engine.rulebook.send(self, i=i, v=None)
		self.send(self, i=i, v=None)


class RuleMapping(MutableMapping, Signal):
	"""Wraps a :class:`RuleBook` so you can get its rules by name.

	You can access the rules in this either dictionary-style or as
	attributes. This is for convenience if you want to get at a rule's
	decorators, eg. to add an Action to the rule.

	Using this as a decorator will create a new rule, named for the
	decorated function, and using the decorated function as the
	initial Action.

	Using this like a dictionary will let you create new rules,
	appending them onto the underlying :class:`RuleBook`; replace one
	rule with another, where the new one will have the same index in
	the :class:`RuleBook` as the old one; and activate or deactivate
	rules. The name of a rule may be used in place of the actual rule,
	so long as the rule already exists.

	:param name: If you want the rule's name to be different from the name
		of its first action, supply the name here.
	:param neighborhood: Optional integer; if supplied, the rule will only
		be run when something's changed within this many nodes.
		``neighborhood=0`` means this only runs when something's changed
		*here*, or a place containing this entity.
	:param big: Set to ``True`` if the rule will make many changes to the world,
		so that Lisien can optimize for a big batch of changes.
	:param always: If set to ``True``, the rule will run every turn.

	"""

	rulebook: RuleBook

	def __init__(self, engine: "Engine", rulebook: RuleBook | RulebookName):
		super().__init__()
		self.engine = engine
		self._rule_cache = self.engine.rule._cache
		if isinstance(rulebook, RuleBook):
			self.rulebook = rulebook
		else:
			self.rulebook = self.engine.rulebook[rulebook]

	def __repr__(self):
		return "RuleMapping({})".format([k for k in self])

	def __iter__(self):
		return iter(self.rulebook._get_cache(*self.engine._btt())[0])

	def __len__(self):
		return len(self.rulebook)

	def __contains__(self, k: Rule | RuleName | str) -> bool:
		return k in self.rulebook

	def __getitem__(self, k: RuleName | str) -> Rule:
		if k not in self:
			raise KeyError("Rule '{}' is not in effect".format(k))
		return self._rule_cache[k]

	def __getattr__(self, k: str) -> Rule:
		if k in self:
			return self[k]
		raise AttributeError

	def __setitem__(
		self, k: RuleName | str, v: Rule | RuleFunc | RuleName | str
	):
		if k == "truth":
			raise KeyError("Illegal rule name")
		if isinstance(v, Hashable) and v in self.engine.rule:
			v = self.engine.rule[v]
		elif isinstance(v, str) and hasattr(self.engine.action, v):
			v = getattr(self.engine.action, v)
		if not isinstance(v, Rule) and callable(v):
			# create a new rule, named k, performing action v
			self.engine.rule[k] = v
			v = self.engine.rule[k]
		assert isinstance(v, Rule)
		if isinstance(k, int):
			self.rulebook[k] = v
		else:
			self.rulebook.append(v)

	def __call__(
		self,
		v: Optional[RuleFunc] = None,
		name: Optional[RuleName | str] = None,
		*,
		neighborhood: Optional[int] = ...,
		big: bool = False,
		always: bool = False,
	):
		def wrap(name, v: RuleFunc, **kwargs):
			name = name if name is not None else v.__name__
			if name == "truth":
				raise ValueError("Illegal rule name")
			self[name] = v
			r = self[name]
			if kwargs.get("always"):
				r.always()
			if "neighborhood" in kwargs:
				r.neighborhood = kwargs["neighborhood"]
			if "big" in kwargs:
				r.big = kwargs["big"]
			return r

		kwargs: dict[str, Any] = {"big": big}
		if always:
			kwargs["always"] = True
		if neighborhood is not ...:
			kwargs["neighborhood"] = neighborhood
		if v is None:
			return partial(wrap, name, **kwargs)
		return wrap(name, v, **kwargs)

	def __delitem__(self, k: RuleName | str):
		i = self.rulebook.index(k)
		del self.rulebook[i]
		self.send(self, key=k, val=None)

	@property
	def priority(self) -> RulebookPriority:
		return self.rulebook.priority

	@priority.setter
	def priority(self, v: RulebookPriority | float):
		self.rulebook.priority = RulebookPriority(v)


class RuleFollower(ABC):
	"""Interface for that which has a rulebook associated, which you can
	get a :class:`RuleMapping` into

	"""

	__slots__ = ()
	engine: AbstractEngine
	_rulebook: RuleBook

	@cached_property
	def _rule_mapping(self) -> RuleMapping:
		return self._get_rule_mapping()

	@property
	def rule(
		self, v: RuleFunc | None = None, name: RuleName | str | None = None
	) -> RuleMapping | partial[Rule]:
		if v is not None:
			return self._rule_mapping(v, name)
		return self._rule_mapping

	@property
	def rulebook(self) -> RuleBook:
		if not hasattr(self, "_rulebook"):
			self._upd_rulebook()
		return self._rulebook

	@rulebook.setter
	def rulebook(self, v: RuleBook | RulebookName | str):
		n = v.name if isinstance(v, RuleBook) else v
		try:
			if n == self._get_rulebook_name():
				return
		except KeyError:
			pass
		self._set_rulebook_name(n)
		self._upd_rulebook()

	def _upd_rulebook(self):
		self._rulebook = self._get_rulebook()

	def _get_rulebook(self) -> RuleBook:
		return self.engine.rulebook[self._get_rulebook_name()]

	def rules(self) -> Iterator[Rule]:
		if not hasattr(self, "engine"):
			raise AttributeError("Need an engine before I can get rules")
		return self._rule_mapping.values()

	@abstractmethod
	def _get_rule_mapping(self):
		"""Get the :class:`RuleMapping` for my rulebook."""
		raise NotImplementedError("_get_rule_mapping")

	@abstractmethod
	def _get_rulebook_name(self):
		"""Get the name of my rulebook."""
		raise NotImplementedError("_get_rulebook_name")

	@abstractmethod
	def _set_rulebook_name(self, n: RulebookName | str):
		"""Tell the database that this is the name of the rulebook to use for
		me.

		"""
		raise NotImplementedError("_set_rulebook_name")


class AllRuleBooks(MutableMapping, Signal):
	__slots__ = ["engine", "_cache"]

	def __init__(self, engine: "Engine"):
		super().__init__()
		self.engine = engine
		self._cache = {}

	def __iter__(self) -> Iterator[RulebookName]:
		return self.engine._rulebooks_cache.iter_entities(*self.engine._btt())

	def __len__(self):
		return len(list(self))

	def __contains__(self, k: RulebookName | Key):
		return self.engine._rulebooks_cache.contains_entity(
			k, *self.engine._btt()
		)

	def __getitem__(self, k: RulebookName | Key) -> RuleBook:
		if k not in self._cache:
			self._cache[k] = RuleBook(self.engine, k)
		return self._cache[k]

	def __setitem__(
		self, key: RulebookName | Key, value: RuleBook | list[Rule | RuleName]
	):
		if key not in self._cache:
			self._cache[key] = RuleBook(self.engine, key)
		rb = self._cache[key]
		while len(rb) > 0:
			del rb[0]
		rb.extend(value)

	def __delitem__(self, key: RulebookName | Key):
		self.engine._del_rulebook(key)


class AllRules(MutableMapping, Signal):
	"""A mapping of every rule in the game.

	You can use this as a decorator to make a rule and not assign it
	to anything.

	"""

	def __init__(self, engine: "Engine"):
		super().__init__()
		self.engine = engine

	@cached_property
	def _cache(self) -> dict[RuleName, Rule]:
		return self.engine._rules_cache

	def __iter__(self) -> Iterator[RuleName]:
		yield from self._cache

	def __len__(self):
		return len(self._cache)

	def __contains__(self, k: RuleName | str):
		return k in self._cache

	def __getitem__(self, k: RuleName | str) -> Rule:
		return self._cache[k]

	def __setitem__(
		self, k: RuleName | str, v: Rule | RuleFunc | RuleFuncName | str
	):
		# you can use the name of a stored function or rule
		if isinstance(v, str):
			if hasattr(self.engine.action, v):
				v = getattr(self.engine.action, v)
			elif hasattr(self.engine.function, v):
				v = getattr(self.engine.function, v)
			elif hasattr(self.engine.rule, v):
				v = getattr(self.engine.rule, v)
			else:
				raise ValueError("Unknown function: " + v)
		if callable(v):
			self._cache[k] = Rule(self.engine, k, actions=[v])
			new = self._cache[k]
		elif isinstance(v, Rule):
			self._cache[k] = v
			new = v
		else:
			raise TypeError(
				"Don't know how to store {} as a rule.".format(type(v))
			)
		self.send(self, key=new, rule=v)

	def __delitem__(self, k: RuleName | Key):
		if k not in self:
			raise KeyError("No such rule")
		for rulebook in self.engine.rulebooks.values():
			try:
				del rulebook[rulebook.index(k)]
			except IndexError:
				pass
		del self._cache[k]
		self.send(self, key=k, rule=None)

	def __call__(
		self,
		v: RuleFunc | RuleFuncName | str | None = None,
		name: RuleName | str | None = None,
		*,
		neighborhood: RuleNeighborhood | int | None = ...,
		always: bool = False,
	):
		def r(name: RuleName | str | None, v: RuleFunc, **kwargs):
			if name is None:
				name = v.__name__
			if name == "truth":
				raise ValueError("Illegal rule name")
			self[name] = v
			ret = self[name]
			if kwargs.get("always"):
				ret.triggers.append("truth")
			if "neighborhood" in kwargs:
				ret.neighborhood = neighborhood
			return ret

		kwargs = {}
		if always:
			kwargs["always"] = True
		if neighborhood is not ...:
			kwargs["neighborhood"] = neighborhood
		if v is None:
			return partial(r, name, **kwargs)
		return r(name, v, **kwargs)

	def new_empty(self, name: RuleName | str) -> Rule:
		"""Make a new rule with no actions or anything, and return it."""
		if name in self:
			raise KeyError("Already have rule {}".format(name))
		new = Rule(self.engine, name)
		self._cache[name] = new
		self.send(self, rule=new, active=True)
		return new



import os
import sys

from . import EngineProcessManager

if __name__ == "__main__":
	if os.path.exists(sys.argv[-1]) and os.path.isfile(sys.argv[-1]):
		mgr = EngineProcessManager()
		eng = mgr.start(replay_file=sys.argv[-1], loglevel="debug")
		mgr.shutdown()



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Wrap a lisien engine so you can access and control it using only
ordinary method calls.

"""

from importlib import import_module
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING, Handler, Logger
from re import match
from typing import Any, Callable, Iterable, Optional

import msgpack
import networkx as nx
import tblib

from ..exc import HistoricKeyError, OutOfTimelineError
from ..node import Node
from ..portal import Portal
from ..types import (
	ActionFuncName,
	Branch,
	CharDelta,
	CharName,
	DeltaDict,
	EdgesDict,
	EdgeValDict,
	EternalKey,
	FuncName,
	FuncStoreName,
	Key,
	Keyframe,
	LinearTime,
	NodeName,
	NodeValDict,
	Plan,
	PrereqFuncName,
	RulebookName,
	RulebookPriority,
	RuleName,
	RuleNeighborhood,
	Stat,
	StatDict,
	Tick,
	Time,
	TriggerFuncName,
	Turn,
	UnitsDict,
	Value,
)
from ..util import (
	EDGE_VAL,
	EDGES,
	ELLIPSIS,
	EMPTY_MAPPING,
	ETERNAL,
	NODE_VAL,
	NODES,
	NONE,
	RULEBOOK,
	RULEBOOKS,
	RULES,
	UNITS,
	UNIVERSAL,
	AbstractCharacter,
	BadTimeException,
	timer,
)

SlightlyPackedDeltaType = dict[
	bytes,
	dict[
		bytes,
		bytes | dict[bytes, bytes | dict[bytes, bytes | dict[bytes, bytes]]],
	],
]
FormerAndCurrentType = tuple[dict[bytes, bytes], dict[bytes, bytes]]


def concat_d(r: dict[bytes, bytes]) -> bytes:
	"""Pack a dictionary of msgpack-encoded keys and values into msgpack bytes"""
	resp = msgpack.Packer().pack_map_header(len(r))
	for k, v in r.items():
		resp += k + v
	return resp


def prepacked(fun: Callable) -> Callable:
	fun.prepacked = True
	return fun


class EngineHandleLogHandler(Handler):
	def __init__(self, level: int, log_queue):
		super().__init__(level)
		self._logq = log_queue

	def emit(self, record):
		if record.exc_info:
			if (
				isinstance(record.exc_info, Exception)
				and record.exc_info.__traceback__
			):
				record.exc_info.__traceback__ = tblib.Traceback(
					record.exc_info.__traceback__
				).as_dict()
			elif (
				isinstance(record.exc_info, tuple)
				and len(record.exc_info) == 3
				and record.exc_info[2]
			):
				record.exc_info = (
					record.exc_info[0],
					record.exc_info[1],
					tblib.Traceback(record.exc_info[2]).as_dict(),
				)
		self._logq.put(record)


class EngineHandle:
	"""A wrapper for a :class:`lisien.Engine` object that runs in the same
	process, but with an API built to be used in a command-processing
	loop that takes commands from another process.

	It's probably a bad idea to use this class unless you're
	developing your own API.

	"""

	_after_ret: Callable

	def __init__(self, *args, log_queue=None, **kwargs):
		"""Instantiate an engine with the given arguments"""
		from ..engine import Engine

		do_game_start = kwargs.pop("do_game_start", False)
		if log_queue:
			logger = kwargs["logger"] = Logger("lisien")
			handler = EngineHandleLogHandler(0, log_queue)
			logger.addHandler(handler)
		self._real = Engine(*args, **kwargs)
		self.pack = pack = self._real.pack

		def pack_pair(pair):
			k, v = pair
			return pack(k), pack(v)

		self.pack_pair = pack_pair
		self.unpack = self._real.unpack

		self._cache_arranger_started = False
		if do_game_start:
			self.do_game_start()

	def log(self, level: str | int, message: str) -> None:
		if isinstance(level, str):
			level = {
				"debug": 10,
				"info": 20,
				"warning": 30,
				"error": 40,
				"critical": 50,
			}[level.lower()]
		self._real.logger.log(level, message)

	def debug(self, message: str) -> None:
		self.log(DEBUG, message)

	def info(self, message: str) -> None:
		self.log(INFO, message)

	def warning(self, message: str) -> None:
		self.log(WARNING, message)

	def error(self, message: str) -> None:
		self.log(ERROR, message)

	def critical(self, message: str) -> None:
		self.log(CRITICAL, message)

	def time_locked(self) -> bool:
		"""Return whether the sim-time has been prevented from advancing"""
		return hasattr(self._real, "locktime")

	def snap_keyframe(self, silent: bool = False) -> Keyframe:
		return self._real.snap_keyframe(silent=silent)

	def _pack_delta(
		self, delta: DeltaDict
	) -> tuple[SlightlyPackedDeltaType, bytes]:
		pack = self.pack
		slightly_packed_delta = {}
		mostly_packed_delta = {}
		for char, chardelta in delta.items():
			if chardelta is ...:
				pchar = pack(char)
				slightly_packed_delta[pchar] = mostly_packed_delta[pchar] = (
					None
				)
				continue
			chardelta = chardelta.copy()
			pchar = pack(char)
			chard = slightly_packed_delta[pchar] = {}
			packd = mostly_packed_delta[pchar] = {}
			if "nodes" in chardelta:
				nd = chard[NODES] = {
					pack(node): pack(ex)
					for node, ex in chardelta.pop("nodes").items()
				}
				packd[NODES] = concat_d(nd)
			if "node_val" in chardelta:
				slightnoded = chard[NODE_VAL] = {}
				packnodevd = {}
				nodedelta: NodeValDict = chardelta.pop("node_val")
				for node, vals in nodedelta.items():
					pnode = pack(node)
					pvals = dict(map(self.pack_pair, vals.items()))
					slightnoded[pnode] = pvals
					packnodevd[pnode] = concat_d(pvals)
				packd[NODE_VAL] = concat_d(packnodevd)
			if "edges" in chardelta:
				edgedelta: EdgesDict = chardelta.pop("edges")
				ed = chard[EDGES] = {
					pack(origdest): pack(ex)
					for origdest, ex in edgedelta.items()
				}
				packd[EDGES] = concat_d(ed)
			if "edge_val" in chardelta:
				slightorigd = chard[EDGE_VAL] = {}
				packorigd = {}
				edgevaldelta: EdgeValDict = chardelta.pop("edge_val")
				for orig, dests in edgevaldelta.items():
					porig = pack(orig)
					slightdestd = slightorigd[porig] = {}
					packdestd = {}
					for dest, port in dests.items():
						pdest = pack(dest)
						slightportd = slightdestd[pdest] = dict(
							map(self.pack_pair, port.items())
						)
						packdestd[pdest] = concat_d(slightportd)
					packorigd[porig] = concat_d(packdestd)
				packd[EDGE_VAL] = concat_d(packorigd)
			if "units" in chardelta:
				slightgraphd = chard[UNITS] = {}
				packunitd = {}
				unitdelta: UnitsDict = chardelta.pop("units")
				for graph, unitss in unitdelta.items():
					if unitss is None:
						continue
					pgraph = pack(graph)
					slightunitd = slightgraphd[pgraph] = dict(
						map(self.pack_pair, unitss.items())
					)
					packunitd[pgraph] = concat_d(slightunitd)
				packd[UNITS] = concat_d(packunitd)
			if "rulebooks" in chardelta:
				chard[RULEBOOKS] = slightrbd = dict(
					map(self.pack_pair, chardelta.pop("rulebooks").items())
				)
				packd[RULEBOOKS] = concat_d(slightrbd)
			todo = dict(map(self.pack_pair, chardelta.items()))
			chard.update(todo)
			packd.update(todo)
		return slightly_packed_delta, concat_d(
			{
				charn: (concat_d(stuff) if stuff is not None else NONE)
				for charn, stuff in mostly_packed_delta.items()
			}
		)

	@staticmethod
	def _concat_char_delta(delta: SlightlyPackedDeltaType) -> bytes:
		delta = delta.copy()
		mostly_packed_delta = packd = {}
		eternal = delta.pop(ETERNAL, None)
		if eternal:
			mostly_packed_delta[ETERNAL] = eternal
		universal = delta.pop(UNIVERSAL, None)
		if universal:
			mostly_packed_delta[UNIVERSAL] = universal
		if RULEBOOK in delta:
			mostly_packed_delta[RULEBOOK] = delta.pop(RULEBOOK)
		if RULES in delta:
			rules = delta.pop(RULES)
			mostly_packed_delta[RULES] = concat_d(
				{rule: concat_d(funcls) for (rule, funcls) in rules.items()}
			)
		if NODES in delta:
			charnodes = delta.pop(NODES)
			packd[NODES] = concat_d(charnodes)
		if NODE_VAL in delta:
			slightnoded = {}
			packnodevd = {}
			for node, vals in delta.pop(NODE_VAL).items():
				slightnoded[node] = vals
				packnodevd[node] = concat_d(vals)
			packd[NODE_VAL] = concat_d(packnodevd)
		if EDGES in delta:
			es = delta.pop(EDGES)
			packd[EDGES] = concat_d(es)
		if EDGE_VAL in delta:
			packorigd = {}
			for orig, dests in delta.pop(EDGE_VAL).items():
				slightdestd = {}
				packdestd = {}
				for dest, port in dests.items():
					slightdestd[dest] = port
					packdestd[dest] = concat_d(port)
				packorigd[orig] = concat_d(packdestd)
			packd[EDGE_VAL] = concat_d(packorigd)
		if UNITS in delta:
			if delta[UNITS] == NONE:
				packd[UNITS] = concat_d({})
				del delta[UNITS]
			else:
				packd[UNITS] = delta.pop(UNITS)
		mostly_packed_delta.update(delta)
		return concat_d(mostly_packed_delta)

	@prepacked
	def next_turn(self) -> tuple[bytes, bytes]:
		"""Simulate a turn. Return whatever result, as well as a delta"""
		pack = self.pack
		self.debug(
			"calling next_turn at {}, {}, {}".format(*self._real._btt())
		)
		ret, delta = self._real.next_turn()
		slightly_packed_delta, packed_delta = self._pack_delta(delta)
		return pack(ret), packed_delta

	def _get_slow_delta(
		self,
		btt_from: Time | None = None,
		btt_to: Time | None = None,
	) -> SlightlyPackedDeltaType:
		return self._real._get_slow_delta(btt_from, btt_to)

	def bookmarks_dump(self) -> list[tuple[Key, Time]]:
		return list(self._real.query.bookmark_items())

	def set_bookmark(self, key: Key, time: Time | None = None) -> Time:
		if time is None:
			self._real.bookmark(key)
			return tuple(self._real.time)
		else:
			self._real.bookmark[key] = time
			return time

	def del_bookmark(self, key: Key):
		del self._real.bookmark[key]

	def start_branch(
		self, parent: Branch, branch: Branch, turn: Turn, tick: Tick
	):
		self._real._start_branch(parent, branch, turn, tick)

	def extend_branch(self, branch: Branch, turn: Turn, tick: Tick):
		self._real._extend_branch(branch, turn, tick)

	def load_at(self, branch: Branch, turn: Turn, tick: Tick):
		self._real.load_at(branch, turn, tick)

	def turn_end(
		self, branch: Branch | None = None, turn: Turn | None = None
	) -> Tick:
		return self._real.turn_end(branch, turn)

	def turn_end_plan(
		self, branch: Branch | None = None, turn: Turn | None = None
	) -> Tick:
		return self._real.turn_end_plan(branch, turn)

	def branch_end(self, branch: Optional[Branch] = None) -> Turn:
		return self._real.branch_end(branch)

	def branch_end_turn_and_tick(self, branch: Branch) -> LinearTime:
		branch = Branch(branch)
		turn = self._real.branch_end_turn(branch)
		return turn, self._real.turn_end(branch, turn)

	def branch_end_plan_turn_and_tick(self, branch: Branch) -> LinearTime:
		branch = Branch(branch)
		turn = self._real.branch_end_turn(branch)
		return turn, self._real.turn_end_plan(branch, turn)

	def start_plan(self) -> Plan:
		self._plan_ctx = self._real.plan()
		return self._plan_ctx.__enter__()

	def end_plan(self) -> tuple[None, DeltaDict]:
		time_was = self._real._btt()
		self._plan_ctx.__exit__(None, None, None)
		del self._plan_ctx
		return None, self._real.get_delta(time_was, self._real._btt())

	@prepacked
	def time_travel(
		self,
		branch: Branch,
		turn: Turn,
		tick: Tick | None = None,
	) -> tuple[bytes, bytes]:
		"""Go to a different `(branch, turn, tick)` and return a delta

		For compatibility with `next_turn` this actually returns a tuple,
		the 0th item of which is `None`.

		"""
		if branch in self._real.branches():
			if self._real._enforce_end_of_time:
				turn_end, tick_end = self._real._branch_end(branch)
				if (tick is None and turn > turn_end) or (
					tick is not None and (turn, tick) > (turn_end, tick_end)
				):
					raise OutOfTimelineError(
						"Not traveling past the end of the branch",
						branch,
						turn,
						tick,
						turn_end,
						tick_end,
					)
			self._real.load_at(branch, turn, tick)
		branch_from, turn_from, tick_from = self._real._btt()
		if tick is None:
			if (
				branch,
				turn,
				self._real.turn_end(branch, turn),
			) == (
				branch_from,
				turn_from,
				tick_from,
			):
				return NONE, EMPTY_MAPPING
			self._real.time = (branch, turn, self._real.turn_end(branch, turn))
		else:
			if (branch, turn, tick) == (
				branch_from,
				turn_from,
				tick_from,
			):
				return NONE, EMPTY_MAPPING
			self._real.time = (branch, turn, tick)
		if turn_from != turn and (
			branch_from != branch
			or None in (turn_from, turn)
			or self._real._is_timespan_too_big(branch, turn_from, turn)
		):
			# This branch avoids unpacking and re-packing the delta
			slightly: SlightlyPackedDeltaType = self._real._get_slow_delta(
				(branch_from, turn_from, tick_from), self._real._btt()
			)
			mostly = {}
			if UNIVERSAL in slightly:
				mostly[UNIVERSAL] = concat_d(slightly.pop(UNIVERSAL))
			if RULES in slightly:
				mostly[RULES] = concat_d(
					{
						rule: concat_d(rule_d)
						for (rule, rule_d) in slightly.pop(RULES).items()
					}
				)
			if RULEBOOK in slightly:
				mostly[RULEBOOK] = concat_d(slightly.pop(RULEBOOK))
			for char, chardeltapacked in slightly.items():
				if chardeltapacked == ELLIPSIS:
					mostly[char] = ELLIPSIS
					continue
				mostly[char] = self._concat_char_delta(chardeltapacked)
			return NONE, concat_d(mostly)
		return NONE, self._pack_delta(
			self._real.get_delta(
				(branch_from, turn_from, tick_from), self._real._btt()
			)
		)[1]

	@prepacked
	def increment_branch(self) -> bytes:
		"""Generate a new branch name and switch to it

		Returns the name of the new branch.

		"""
		branch = self._real.branch
		m = match(r"(.*)(\d+)", branch)
		if m:
			stem, n = m.groups()
			branch = stem + str(int(n) + 1)
		else:
			stem = branch
			n = 1
			branch = stem + str(n)
		if branch in self._real.branches():
			if m:
				n = int(n)
			else:
				stem = branch[:-1]
				n = 1
			while stem + str(n) in self._real.branches():
				n += 1
			branch = stem + str(n)
		self._real.branch = branch
		return self.pack(branch)

	def add_character(
		self,
		char: CharName,
		data: nx.Graph | nx.DiGraph | None = None,
		node: NodeValDict | None = None,
		edge: EdgeValDict | None = None,
		**attr,
	):
		"""Make a new character, initialized with whatever data"""
		# Probably not great that I am unpacking and then repacking the stats
		self._real.add_character(char, data=data, node=node, edge=edge, **attr)

	def commit(self):
		self._real.commit()

	def close(self):
		self._real.close()

	def get_btt(self) -> Time:
		return self._real._btt()

	def get_language(self) -> str:
		return str(self._real.string.language)

	def set_language(self, lang: str) -> dict[str, str]:
		self._real.string.language = lang
		return self.strings_copy(lang)

	def get_string_lang_items(
		self, lang: str | None = None
	) -> list[tuple[str, str]]:
		return list(self._real.string.lang_items(lang))

	def strings_copy(self, lang: str | None = None) -> dict[str, str]:
		return dict(self._real.string.lang_items(lang))

	def set_string(self, k: str, v: str) -> None:
		self._real.string[k] = v

	def del_string(self, k: str) -> None:
		del self._real.string[k]

	@prepacked
	def get_eternal(self, k: EternalKey) -> bytes:
		return self.pack(self._real.eternal[k])

	def set_eternal(self, k: EternalKey, v: Value) -> None:
		self._real.eternal[k] = v

	def del_eternal(self, k: EternalKey) -> None:
		del self._real.eternal[k]

	@prepacked
	def eternal_copy(self) -> dict[bytes, bytes]:
		return dict(map(self.pack_pair, self._real.eternal.items()))

	def set_universal(self, k: EternalKey, v: Value) -> None:
		self._real.universal[k] = v

	def del_universal(self, k: EternalKey) -> None:
		del self._real.universal[k]

	def del_character(self, char: CharName) -> None:
		del self._real.character[char]

	def set_character(self, char: CharName, data: CharDelta) -> None:
		self._real.character[char] = data

	def set_character_stat(self, char: CharName, k: Stat, v: Value) -> None:
		self._real.character[char].stat[k] = v

	def del_character_stat(self, char: CharName, k: Stat) -> None:
		del self._real.character[char].stat[k]

	def set_node_stat(
		self, char: CharName, node: NodeName, k: Stat, v: Value
	) -> None:
		self._real.character[char].node[node][k] = v

	def del_node_stat(self, char: CharName, node: NodeName, k: Stat) -> None:
		del self._real.character[char].node[node][k]

	def _get_btt(self, btt: Time | None = None) -> Time:
		if btt is None:
			return self._real._btt()
		return btt

	def node_exists(self, char: CharName, node: NodeName) -> bool:
		return self._real._node_exists(char, node)

	def update_nodes(self, char: CharName, patch: NodeValDict):
		"""Change the stats of nodes in a character according to a
		dictionary.

		"""
		node = self._real.character[char].node
		with (
			self._real.batch(),
			timer("EngineHandle.update_nodes", self.debug),
		):
			for n, npatch in patch.items():
				if npatch is None:
					del node[n]
				elif n not in node:
					node[n] = npatch
				else:
					node[n].update(npatch)

	def del_node(self, char: CharName, node: NodeName):
		"""Remove a node from a character."""
		del self._real.character[char].node[node]

	def character_set_node_predecessors(
		self, char: CharName, node: NodeName, preds: Iterable
	) -> None:
		self._real.character[char].pred[node] = preds

	def set_thing(
		self, char: CharName, thing: NodeName, statdict: dict
	) -> None:
		self._real.character[char].thing[thing] = statdict

	def add_thing(
		self, char: CharName, thing: NodeName, loc: NodeName, statdict: dict
	) -> None:
		self._real.character[char].add_thing(thing, loc, **statdict)

	def place2thing(self, char: CharName, place: NodeName, loc: NodeName):
		self._real.character[char].place2thing(place, loc)

	def thing2place(self, char: CharName, thing: NodeName):
		self._real.character[char].thing2place(thing)

	def set_thing_location(
		self, char: CharName, thing: NodeName, loc: NodeName
	) -> None:
		self._real.character[char].thing[thing]["location"] = loc

	def thing_follow_path(
		self,
		char: CharName,
		thing: NodeName,
		path: list[NodeName],
		weight: Stat,
	) -> int:
		return (
			self._real.character[char].thing[thing].follow_path(path, weight)
		)

	def thing_go_to_place(
		self, char: CharName, thing: NodeName, place: NodeName, weight: Stat
	) -> int:
		return (
			self._real.character[char].thing[thing].go_to_place(place, weight)
		)

	def thing_travel_to(
		self,
		char: CharName,
		thing: NodeName,
		dest: NodeName,
		weight: Stat | None = None,
		graph: nx.DiGraph | None = None,
	) -> int:
		"""Make something find a path to ``dest`` and follow it.

		Optional argument ``weight`` is the portal stat to use to schedule
		movement times.

		Optional argument ``graph`` is an alternative graph to use for
		pathfinding. Should resemble a networkx DiGraph.

		"""
		return (
			self._real.character[char]
			.thing[thing]
			.travel_to(dest, weight, graph)
		)

	def set_place(
		self, char: CharName, place: NodeName, statdict: StatDict
	) -> None:
		self._real.character[char].place[place] = statdict

	def add_places_from(self, char: CharName, seq: Iterable) -> None:
		self._real.character[char].add_places_from(seq)

	def add_portal(
		self,
		char: CharName,
		orig: NodeName,
		dest: NodeName,
		statdict: StatDict,
		symmetrical: bool = False,
	) -> None:
		self._real.character[char].add_portal(orig, dest, **statdict)
		if symmetrical:
			self._real.character[char].add_portal(dest, orig, **statdict)

	def add_portals_from(self, char: CharName, seq: Iterable) -> None:
		self._real.character[char].add_portals_from(seq)

	def del_portal(
		self, char: CharName, orig: NodeName, dest: NodeName
	) -> None:
		ch = self._real.character[char]
		ch.remove_edge(orig, dest)
		assert orig in ch.node
		assert dest in ch.node

	def set_portal(
		self, char: CharName, orig: NodeName, dest: NodeName, v: StatDict
	):
		self._real.character[CharName(char)].portal[NodeName(orig)][
			NodeName(dest)
		] = v

	def set_portal_stat(
		self, char: CharName, orig: NodeName, dest: NodeName, k: Stat, v: Value
	) -> None:
		self._real.character[char].portal[orig][dest][k] = v

	def del_portal_stat(
		self, char: CharName, orig: NodeName, dest: NodeName, k: Stat
	) -> None:
		del self._real.character[char][orig][dest][k]

	def add_unit(
		self, char: CharName, graph: CharName, node: NodeName
	) -> None:
		self._real.character[char].add_unit(graph, node)

	def remove_unit(
		self, char: CharName, graph: CharName, node: NodeName
	) -> None:
		self._real.character[char].remove_unit(graph, node)

	def new_empty_rule(self, rule: RuleName) -> None:
		self._real.rule.new_empty(rule)

	def new_empty_rulebook(self, rulebook: RulebookName) -> list:
		self._real.rulebook.__getitem__(rulebook)
		return []

	def set_rulebook_priority(
		self, rulebook: RulebookName, priority: RulebookPriority
	) -> None:
		self._real.rulebook[rulebook].priority = priority

	def set_rulebook_rules(
		self, rulebook: RulebookName, rules: list[RuleName]
	) -> None:
		self._real.rulebook[rulebook] = rules

	def set_rulebook_rule(
		self, rulebook: RulebookName, i: int, rule: RuleName
	) -> None:
		self._real.rulebook[rulebook][i] = rule

	def ins_rulebook_rule(
		self, rulebook: RulebookName, i: int, rule: RuleName
	) -> None:
		self._real.rulebook[rulebook].insert(i, rule)

	def del_rulebook_rule(self, rulebook: RulebookName, i: int) -> None:
		del self._real.rulebook[rulebook][i]

	def move_rulebook_rule_back(self, rulebook: RulebookName, i: int) -> int:
		rb = self._real.rulebook[rulebook]
		rule = rb.pop(i)
		rb.insert(i - 1, rule)
		return i - 1

	def move_rulebook_rule_forward(
		self, rulebook: RulebookName, i: int
	) -> int:
		rb = self._real.rulebook[rulebook]
		rule = rb.pop(i)
		rb.insert(i + 1, rule)
		return i + 1

	def del_rulebook(self, rulebook: RulebookName) -> None:
		del self._real.rulebook[rulebook]

	def del_rule(self, rule: RuleName) -> None:
		del self._real.rule[rule]

	def set_rule_triggers(
		self, rule: RuleName, triggers: list[TriggerFuncName]
	) -> None:
		self._real.rule[rule].triggers = triggers

	def set_rule_prereqs(
		self, rule: RuleName, prereqs: list[PrereqFuncName]
	) -> None:
		self._real.rule[rule].prereqs = prereqs

	def set_rule_actions(
		self, rule: RuleName, actions: list[ActionFuncName]
	) -> None:
		self._real.rule[rule].actions = actions

	def set_rule_neighborhood(
		self, rule: RuleName, neighborhood: RuleNeighborhood
	) -> None:
		self._real.rule[rule].neighborhood = neighborhood

	def get_rule_neighborhood(self, rule: RuleName) -> RuleNeighborhood:
		return self._real.rule[rule].neighborhood

	def set_character_rulebook(
		self, char: CharName, rulebook: RulebookName
	) -> None:
		self._real.character[char].rulebook = rulebook

	def set_unit_rulebook(
		self, char: CharName, rulebook: RulebookName
	) -> None:
		self._real.character[char].unit.rulebook = rulebook

	def set_character_thing_rulebook(
		self, char: CharName, rulebook: RulebookName
	) -> None:
		self._real.character[char].thing.rulebook = rulebook

	def set_character_place_rulebook(
		self, char: CharName, rulebook: RulebookName
	) -> None:
		self._real.character[char].place.rulebook = rulebook

	def set_character_node_rulebook(
		self, char: CharName, rulebook: RulebookName
	) -> None:
		self._real.character[char].node.rulebook = rulebook

	def set_character_portal_rulebook(
		self, char: CharName, rulebook: RulebookName
	) -> None:
		self._real.character[char].portal.rulebook = rulebook

	def set_node_rulebook(
		self, char: CharName, node: NodeName, rulebook: RulebookName
	) -> None:
		self._real.character[char].node[node].rulebook = rulebook

	def set_portal_rulebook(
		self,
		char: CharName,
		orig: NodeName,
		dest: NodeName,
		rulebook: RulebookName,
	) -> None:
		self._real.character[char].portal[orig][dest].rulebook = rulebook

	@prepacked
	def source_copy(
		self,
		store: FuncStoreName,
	) -> dict[bytes, bytes]:
		return dict(
			map(self.pack_pair, getattr(self._real, store).iterplain())
		)

	def get_source(self, store: FuncStoreName, name: FuncName) -> str:
		return getattr(self._real, store).get_source(name)

	def store_source(
		self, store: FuncStoreName, v: str, name: FuncName | None = None
	) -> None:
		getattr(self._real, store).store_source(v, name)

	def del_source(self, store: FuncStoreName, k: FuncName) -> None:
		delattr(getattr(self._real, store), k)

	def call_stored_function(
		self, store: FuncStoreName, func: FuncName, args: tuple, kwargs: dict
	) -> Any:
		branch, turn, tick = self._real._btt()
		if store == "method":
			args = (self._real,) + tuple(args)
		store = getattr(self._real, store)
		if store not in self._real.stores:
			raise ValueError("{} is not a function store".format(store))
		callme = getattr(store, func)
		res = callme(*args, **kwargs)
		_, turn_now, tick_now = self._real._btt()
		delta = self._real._get_branch_delta(
			branch, turn, tick, turn_now, tick_now
		)
		return res, delta

	def call_randomizer(self, method: str, *args, **kwargs) -> Any:
		return getattr(self._real._rando, method)(*args, **kwargs)

	def install_module(self, module: str) -> None:
		import_module(module).install(self._real)

	def do_game_start(self):
		branch, turn, tick = self._real._btt()
		self._real.game_start()
		return [], self._real._get_branch_delta(
			branch, turn, tick, self._real.turn, self._real.tick
		)

	def is_ancestor_of(self, parent: Branch, child: Branch) -> bool:
		return self._real.is_ancestor_of(parent, child)

	def branch_start(self, branch: Branch) -> LinearTime:
		return self._real._branch_start(branch)

	def branch_parent(self, branch: Branch) -> str | None:
		return self._real.branch_parent(branch)

	def apply_choices(
		self,
		choices: list[dict],
		dry_run: bool = False,
		perfectionist: bool = False,
	) -> tuple[list[tuple[Any, Any]], list[tuple[Any, Any]]]:
		return self._real.apply_choices(choices, dry_run, perfectionist)

	@staticmethod
	def get_schedule(
		entity: AbstractCharacter | Node | Portal,
		stats: Iterable[Key],
		beginning: int,
		end: int,
	) -> dict[Key, list]:
		ret = {}
		for stat in stats:
			ret[stat] = list(
				entity.historical(stat).iter_history(beginning, end)
			)
		return ret

	def rules_handled_turn(
		self, branch: Branch | None = None, turn: Turn | None = None
	) -> dict[str, dict[int, dict[int, str]]]:
		if branch is None:
			branch = self._real.branch
		if turn is None:
			turn = self._real.turn
		eng = self._real
		# assume the caches are all sync'd
		try:
			return {
				"character": eng._character_rules_handled_cache.handled_deep[
					branch
				][turn],
				"unit": eng._unit_rules_handled_cache.handled_deep[branch][
					turn
				],
				"character_thing": eng._character_thing_rules_handled_cache.handled_deep[
					branch
				][turn],
				"character_place": eng._character_place_rules_handled_cache.handled_deep[
					branch
				][turn],
				"character_portal": eng._character_portal_rules_handled_cache.handled_deep[
					branch
				][turn],
				"node": eng._node_rules_handled_cache.handled_deep[branch][
					turn
				],
				"portal": eng._portal_rules_handled_cache.handled_deep[branch][
					turn
				],
			}
		except HistoricKeyError:
			return {
				"character": {},
				"unit": {},
				"character_thing": {},
				"character_place": {},
				"character_portal": {},
				"node": {},
				"portal": {},
			}

	def branches(
		self,
	) -> dict[Branch, tuple[Branch | None, Turn, Tick, Turn, Tick]]:
		return dict(self._real._branches_d)

	def main_branch(self) -> Branch:
		return self._real.main_branch

	def switch_main_branch(self, branch: Branch) -> Keyframe:
		self._real.switch_main_branch(branch)
		return self.snap_keyframe()

	def game_init(
		self,
	) -> tuple[
		Keyframe,
		dict[EternalKey, Value],
		dict[FuncName, str],
		dict[FuncName, str],
		dict[TriggerFuncName, str],
		dict[PrereqFuncName, str],
		dict[ActionFuncName, str],
	]:
		branch, turn, tick = self._real._btt()
		if (turn, tick) != (0, 0):
			raise BadTimeException(
				"You tried to start a game when it wasn't the start of time"
			)
		self.do_game_start()
		kf = self.snap_keyframe()
		functions = dict(self._real.function.iterplain())
		methods = dict(self._real.method.iterplain())
		triggers = dict(self._real.trigger.iterplain())
		prereqs = dict(self._real.prereq.iterplain())
		actions = dict(self._real.action.iterplain())
		return (
			kf,
			self._real.eternal.copy(),
			functions,
			methods,
			triggers,
			prereqs,
			actions,
		)


def serial_handle(prefix, **kwargs):
	kwargs["workers"] = 0
	return EngineHandle(prefix, **kwargs)



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The nodes of lisien's character graphs.

Every node that actually exists is either a Place or a Thing, but they
have a lot in common.

"""

from __future__ import annotations

from collections.abc import Mapping, Set, ValuesView
from typing import TYPE_CHECKING, Iterator, List, Literal, Optional

from networkx import shortest_path, shortest_path_length

from . import graph, rule
from .exc import AmbiguousLeaderError
from .facade import FacadePlace, FacadeThing
from .query import EntityStatAlias
from .rule import RuleMapping
from .types import (
	CharName,
	Key,
	NodeName,
	RulebookName,
	Stat,
	Time,
	Turn,
	Value,
)
from .util import AbstractThing, getatt

if TYPE_CHECKING:
	from .character import Character
	from .engine import Engine
	from .portal import Portal


class LeaderMapping(Mapping):
	"""A mapping of the characters that have a particular node as a unit.

	Getting characters from here isn't any better than getting them from
	the engine direct, but with this you can do things like use the
	.get() method to get a character if it's a user and otherwise
	get something else; or test whether the character's name is in
	the keys; and so on.

	"""

	__slots__ = ["node"]

	def __init__(self, node) -> None:
		"""Store the node"""
		self.node = node

	engine: "Engine" = getatt("node.engine")

	def _user_names(self) -> frozenset[CharName]:
		try:
			return self.engine._unitness_cache.leader_cache.retrieve(
				self.node.character.name, self.node.name, *self.engine._btt()
			)
		except KeyError:
			return frozenset()

	@property
	def only(self) -> Character:
		"""If there's only one leader, return it.

		Otherwise, raise ``AmbiguousLeaderError``, a type of ``AttributeError``.

		"""
		user_names = self._user_names()
		if len(user_names) != 1:
			raise AmbiguousLeaderError(
				"No leaders, or more than one",
				self.node.name,
				*self.engine._btt(),
				user_names,
			)
		user_name = next(iter(user_names))
		return self.engine.character[user_name]

	def __iter__(self) -> Iterator[CharName]:
		yield from self._user_names()

	def __contains__(self, item: CharName) -> bool:
		return item in self.engine._unitness_cache.leader_cache.retrieve(
			self.node.character.name, self.node.name, *self.engine._btt()
		)

	def __len__(self) -> int:
		return len(self._user_names())

	def __bool__(self) -> bool:
		for _ in self._user_names():
			return True
		return False

	def __getitem__(self, k: CharName) -> "Character":
		ret = self.engine.character[k]
		node = self.node
		charn = node.character.name
		nn = node.name
		avatar = ret.unit
		if charn not in avatar or nn not in avatar[charn]:
			raise KeyError(
				"{} not used by {}".format(self.node.name, k),
				self.engine._btt(),
			)
		return ret


class NodeContentValues(ValuesView):
	_mapping: "NodeContent"

	def __iter__(self) -> Iterator["Thing"]:
		node = self._mapping.node
		nodem = node.character.node
		try:
			conts = node.engine._node_contents(node.character.name, node.name)
		except KeyError:
			return
		for name in conts:
			if name not in nodem:
				return
			yield nodem[name]

	def __contains__(self, item: Thing) -> bool:
		try:
			return item.location == self._mapping.node
		except AttributeError:
			return False


class NodeContent(Mapping):
	__slots__ = ("node",)

	def __init__(self, node: Node) -> None:
		self.node = node

	def __iter__(self) -> Iterator[NodeName]:
		try:
			it = self.node.engine._node_contents_cache.retrieve(
				self.node.character.name,
				self.node.name,
				*self.node.engine._btt(),
			)
		except KeyError:
			return
		yield from it

	def __len__(self) -> int:
		try:
			return len(
				self.node.engine._node_contents_cache.retrieve(
					self.node.character.name,
					self.node.name,
					*self.node.engine._btt(),
				)
			)
		except KeyError:
			return 0

	def __contains__(self, item: NodeName) -> bool:
		try:
			return self.node.character.thing[item].location == self.node
		except KeyError:
			return False

	def __getitem__(self, item: NodeName) -> "Thing":
		if item not in self:
			raise KeyError
		return self.node.character.thing[item]

	def values(self) -> NodeContentValues:
		return NodeContentValues(self)


class DestsValues(ValuesView):
	_mapping: "Dests"

	def __contains__(self, item: "Portal") -> bool:
		_, name = self._mapping._pn
		return item.origin.name == name


class Dests(Mapping):
	__slots__ = ("_ecnb", "_pn")

	def __init__(self, node: Node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self) -> Iterator[NodeName]:
		edges_cache, charname, name, btt = self._ecnb
		for succ in edges_cache.iter_successors(charname, name, *btt()):
			if succ in self:
				yield succ

	def __len__(self) -> int:
		n = 0
		for n, _ in enumerate(self, start=1):
			pass
		return n

	def __contains__(self, item) -> bool:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_successor(charname, name, item, *btt())

	def __getitem__(self, item) -> "Portal":
		portal, name = self._pn
		return portal[name][item]

	def values(self) -> DestsValues:
		return DestsValues(self)


class OrigsValues(ValuesView):
	_mapping: "Origs"

	def __contains__(self, item: "Portal") -> bool:
		_, name = self._mapping._pn
		return item.destination.name == name


class Origs(Mapping):
	__slots__ = ("_pn", "_ecnb")

	def __init__(self, node: Node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self) -> Iterator[NodeName]:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.iter_predecessors(charname, name, *btt())

	def __contains__(self, item: NodeName) -> bool:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_predecessor(charname, name, item, *btt())

	def __len__(self) -> int:
		edges_cache, charname, name, btt = self._ecnb
		n = 0
		for n, _ in enumerate(self, start=1):
			pass
		return n

	def __getitem__(self, item: NodeName) -> "Node":
		if item not in self:
			raise KeyError
		portal, name = self._pn
		return portal[item][name]

	def values(self) -> OrigsValues:
		return OrigsValues(self)


class Portals(Set):
	__slots__ = ("_pn", "_pecnb")

	def __init__(self, node: Node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._pecnb = (
			engine._get_edge,
			engine._edges_cache,
			character,
			character.name,
			name,
			engine._btt,
		)

	def __contains__(self, x: NodeName) -> bool:
		_, edges_cache, _, charname, name, btt_f = self._pecnb
		btt = btt_f()
		return edges_cache.has_predecessor(
			charname, name, x, *btt
		) or edges_cache.has_successor(charname, name, x, *btt)

	def __len__(self) -> int:
		_, edges_cache, _, charname, name, btt_f = self._pecnb
		btt = btt_f()
		stuff = set()
		for pred in edges_cache.iter_predecessors(charname, name, *btt):
			if edges_cache.has_predecessor(charname, name, pred, *btt):
				stuff.add((pred, name))
		for succ in edges_cache.iter_successors(charname, name, *btt):
			if edges_cache.has_successor(charname, name, succ, *btt):
				stuff.add((name, succ))
		return len(stuff)

	def __iter__(self) -> Iterator["Portal"]:
		get_edge, edges_cache, character, charname, name, btt_f = self._pecnb
		btt = btt_f()
		for dest in edges_cache.iter_successors(charname, name, *btt):
			if edges_cache.has_successor(charname, name, dest, *btt):
				yield get_edge(character, name, dest)
		for orig in edges_cache.iter_predecessors(charname, name, *btt):
			if edges_cache.has_predecessor(charname, name, orig, *btt):
				yield get_edge(character, orig, name)


class NeighborValues(ValuesView):
	_mapping: "NeighborMapping"

	def __contains__(self, item: Node) -> bool:
		return item.name in self._mapping


class NeighborMapping(Mapping):
	__slots__ = ("_nn", "_ecnb")

	def __init__(self, node: "Node") -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._nn = (character.node, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self) -> Iterator[NodeName]:
		edges_cache, charname, name, btt = self._ecnb
		seen = set()
		for succ in edges_cache.iter_successors(charname, name, *btt()):
			yield succ
			seen.add(succ)
		for pred in edges_cache.iter_predecessors(charname, name, *btt()):
			if pred in seen:
				continue
			yield pred
			seen.add(pred)

	def __contains__(self, item: NodeName) -> bool:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_predecessor(
			charname, name, item, *btt()
		) or edges_cache.has_successor(charname, name, item, *btt())

	def __len__(self) -> int:
		return len(set(iter(self)))

	def __getitem__(self, item: NodeName) -> "Node":
		node, name = self._nn
		if item not in self:
			raise KeyError(f"{item} is not a neighbor of {name}")
		return node[item]

	def values(self) -> NeighborValues:
		return NeighborValues(self)


class Node(graph.Node, rule.RuleFollower):
	"""The fundamental graph component, which portals go between.

	Every lisien node is either a thing or a place. They share in common
	the abilities to follow rules; to be connected by portals; and to
	contain things.

	This is truthy if it exists, falsy if it's been deleted.

	"""

	__slots__ = ("_real_rule_mapping",)
	character = getatt("graph")
	no_unwrap = True
	_extra_keys = {
		"name",
	}

	def _get_rule_mapping(self) -> RuleMapping:
		return rule.RuleMapping(self.db, self.rulebook)

	def _get_rulebook_name(self) -> RulebookName:
		now = self.engine._btt()
		try:
			return self.engine._nodes_rulebooks_cache.retrieve(
				self.character.name, self.name, *now
			)
		except KeyError:
			ret = RulebookName(Key((self.character.name, self.name)))
			self.engine._nodes_rulebooks_cache.store(
				self.character.name, self.name, *now, ret
			)
			self.engine.query.set_node_rulebook(
				self.character.name, self.name, *now, ret
			)
			return ret

	def _get_rulebook(self) -> rule.RuleBook:
		return rule.RuleBook(self.engine, self._get_rulebook_name())

	def _set_rulebook_name(self, rulebook: RulebookName) -> None:
		character = self.character.name
		node = self.name
		cache = self.engine._nodes_rulebooks_cache
		try:
			if rulebook == cache.retrieve(
				character, node, *self.engine._btt()
			):
				return
		except KeyError:
			pass
		branch, turn, tick = self.engine._nbtt()
		cache.store(character, node, branch, turn, tick, rulebook)
		self.engine.query.set_node_rulebook(
			character, node, branch, turn, tick, rulebook
		)

	successor = succ = adj = edge = getatt("portal")
	predecessor = pred = getatt("preportal")
	engine: Engine = getatt("db")

	@property
	def leader(self) -> LeaderMapping:
		__doc__ = LeaderMapping.__doc__
		return LeaderMapping(self)

	def leaders(self) -> ValuesView[Character]:
		return self.leader.values()

	def __init__(self, character: "Character", name: NodeName):
		super().__init__(character, name)
		self.db = getattr(character, "engine", character.db)

	@property
	def neighbor(self) -> NeighborMapping:
		return NeighborMapping(self)

	def neighbors(self):
		return self.neighbor.values()

	@property
	def portal(self) -> Dests:
		"""A mapping of portals leading out from this node.

		Aliases ``portal``, ``adj``, ``edge``, ``successor``, and ``succ``
		are available.

		"""
		return Dests(self)

	@property
	def preportal(self) -> Origs:
		"""A mapping of portals leading to this node.

		Aliases ``preportal``, ``predecessor`` and ``pred`` are available.

		"""
		return Origs(self)

	def portals(self) -> Portals:
		"""A set-like object of portals connected to this node."""
		return Portals(self)

	@property
	def content(self) -> NodeContent:
		"""A mapping of ``Thing`` objects that are here"""
		return NodeContent(self)

	def contents(self) -> NodeContentValues:
		"""A set-like object containing ``Thing`` objects that are here"""
		return self.content.values()

	def clear(self) -> None:
		"""Delete all my keys"""
		for key in super().__iter__():
			del self[key]

	def __contains__(self, k: Stat):
		"""Handle extra keys, then delegate."""
		return k in self._extra_keys or super().__contains__(k)

	def __setitem__(
		self, k: Stat | Literal["rulebook"], v: Value | RulebookName
	):
		if k == "rulebook":
			self._set_rulebook_name(v)
		else:
			super().__setitem__(k, v)

	def __delitem__(self, k: Stat):
		super().__delitem__(k)

	def successors(self) -> Iterator[Node]:
		"""Iterate over nodes with edges leading from here to there."""
		for port in self.portal.values():
			yield port.destination

	def predecessors(self) -> Iterator[Node]:
		"""Iterate over nodes with edges leading here from there."""
		for port in self.preportal.values():
			yield port.origin

	def _plain_dest_name(self, dest: Node | NodeName) -> NodeName:
		if isinstance(dest, Node):
			if dest.character != self.character:
				raise ValueError(
					"{} not in {}".format(dest.name, self.character.name)
				)
			return dest.name
		else:
			if dest in self.character.node:
				return dest
			raise ValueError("{} not in {}".format(dest, self.character.name))

	def shortest_path_length(
		self, dest: NodeName | Node, weight: Stat | None = None
	) -> int:
		"""Return the length of the path from me to ``dest``.

		Raise ``ValueError`` if ``dest`` is not a node in my character
		or the name of one.

		"""

		return shortest_path_length(
			self.character, self.name, self._plain_dest_name(dest), weight
		)

	def shortest_path(
		self, dest: NodeName | Node, weight: Stat | None = None
	) -> List[Key]:
		"""Return a list of node names leading from me to ``dest``.

		Raise ``ValueError`` if ``dest`` is not a node in my character
		or the name of one.

		"""
		return shortest_path(
			self.character, self.name, self._plain_dest_name(dest), weight
		)

	def path_exists(
		self, dest: NodeName | Node, weight: Stat | None = None
	) -> bool:
		"""Return whether there is a path leading from me to ``dest``.

		With ``weight``, only consider edges that have a stat by the
		given name.

		Raise ``ValueError`` if ``dest`` is not a node in my character
		or the name of one.

		"""
		try:
			return bool(self.shortest_path_length(dest, weight))
		except KeyError:
			return False

	def delete(self) -> None:
		"""Get rid of this node

		Apart from deleting the node, this also informs all its users
		that it doesn't exist and therefore can't be their unit
		anymore.

		"""
		self._delete()

	def _delete(self, *, now: Optional[Time] = None) -> Time:
		engine = self.engine
		with (
			engine.world_lock,
			engine.batch(),
			engine._nodes_cache.overwriting(),
			engine._node_val_cache.overwriting(),
			engine._unitness_cache.overwriting(),
		):
			if now is None:
				now = engine._nbtt()
			character = self.character
			g = character.name
			n = self.name
			for contained in list(self.contents()):
				contained._delete(now)
				now = engine._nbtt()
			if n in character.portal:
				for port in list(character.portal[n].values()):
					port._delete(now=now)
					now = engine._nbtt()
			if n in character.preportal:
				for port in list(character.preportal[n].values()):
					port._delete(now=now)
					now = engine._nbtt()
			for username in list(self.leader):
				engine._unitness_cache.store(username, g, n, *now, False)
				engine.query.unit_set(username, g, n, *now, False)
			for k in self:
				assert k != "name"
				if k != "location":
					self._del_cache(k, *now)
					self._del_db(k, *now)
			engine._nodes_cache.store(g, n, *now, False)
			engine.query.exist_node(g, n, *now, False)
			self.character.node.send(
				self.character.node, key=self.name, val=None
			)
			return now

	def add_portal(self, other: NodeName | Node, **stats) -> None:
		"""Connect a portal from here to another node"""
		self.character.add_portal(
			self.name, getattr(other, "name", other), **stats
		)

	def new_portal(self, other: NodeName | Node, **stats) -> "Portal":
		"""Connect a portal from here to another node, and return it."""
		return self.character.new_portal(
			self.name, getattr(other, "name", other), **stats
		)

	def add_thing(self, name: NodeName, **stats) -> None:
		"""Make a new Thing here"""
		self.character.add_thing(name, self.name, **stats)

	def new_thing(self, name: NodeName, **stats) -> "Thing":
		"""Create a new thing, located here, and return it."""
		return self.character.new_thing(name, self.name, **stats)

	def historical(self, stat: Stat) -> EntityStatAlias:
		"""Return a reference to the values that a stat has had in the past.

		You can use the reference in comparisons to make a history
		query, and execute the query by calling it, or passing it to
		``self.engine.ticks_when``.

		"""
		return EntityStatAlias(entity=self, stat=stat)

	def __bool__(self):
		return self.engine._node_exists(self.character.name, self.name)


class Place(Node):
	"""The kind of node where a thing might ultimately be located.

	lisien entities are truthy so long as they exist, falsy if they've
	been deleted.

	"""

	__slots__ = (
		"graph",
		"db",
		"node",
		"_rulebook",
		"_rulebooks",
		"_real_rule_mapping",
	)

	extrakeys = {
		"name",
	}

	def __getitem__(self, key: Stat):
		if key == "name":
			return self.name
		return super().__getitem__(key)

	def __repr__(self):
		return "<{}.character[{}].place[{}]>".format(
			repr(self.engine), self.character.name, self.name
		)

	def _validate_node_type(self) -> bool:
		try:
			self.engine._things_cache.retrieve(
				self.character.name, self.name, *self.engine._btt()
			)
			return False
		except:
			return True

	def facade(self) -> FacadePlace:
		return FacadePlace(self.character.facade(), self)


def roerror(*args):
	raise RuntimeError("Read-only")


class Thing(Node, AbstractThing):
	"""The sort of item that has a particular location at any given time.

	Things are always in Places or other Things, and may additionally be
	travelling through a Portal.

	lisien entities are truthy so long as they exist, falsy if they've
	been deleted.

	"""

	__slots__ = (
		"graph",
		"db",
		"node",
		"_rulebook",
		"_rulebooks",
		"_real_rule_mapping",
	)

	_extra_keys = {"name", "location"}

	def _getname(self) -> NodeName:
		return self.name

	def _getloc(self) -> NodeName | None:
		ret = self.engine._things_cache._base_retrieve(
			(self.character.name, self.name, *self.engine._btt())
		)
		if ret is ... or isinstance(ret, Exception):
			return None
		return ret

	def _validate_node_type(self) -> bool:
		return self._getloc() is not None

	def _get_arrival_time(self) -> Turn:
		charn = self.character.name
		n = self.name
		thingcache = self.engine._things_cache
		for b, trn, tck in self.engine._iter_parent_btt():
			try:
				v = thingcache.turn_before(charn, n, b, trn)
			except KeyError:
				v = thingcache.turn_after(charn, n, b, trn)
			if v is not None:
				return v
		else:
			raise ValueError("Couldn't find arrival time")

	def _set_loc(self, loc: Optional[NodeName]) -> None:
		self.engine._set_thing_loc(self.character.name, self.name, loc)

	def __getitem__(
		self, item: Stat | Literal["location"]
	) -> Value | NodeName:
		if item == "location":
			return self._getloc()
		return super().__getitem__(item)

	def __setitem__(
		self, key: Stat | Literal["location"], value: Value | NodeName
	):
		"""Set ``key``=``value`` for the present game-time."""
		if key == "name":
			raise RuntimeError("Read-only name")
		elif key == "location":
			self._set_loc(NodeName(Key(value)))
		else:
			super().__setitem__(key, value)

	def __delitem__(self, key: Stat):
		"""As of now, this key isn't mine."""
		if key in self._extra_keys:
			raise ValueError("Can't delete {}".format(key))
		super().__delitem__(key)

	def __repr__(self):
		charn = self.character.name
		return f"<{self.engine}.character[{charn}].thing[{self.name}]"

	def facade(self) -> FacadeThing:
		return FacadeThing(self.character.facade(), self)

	def _delete(self, now: Optional[Time] = None) -> Time:
		with (
			self.engine.world_lock,
			self.engine.batch(),
			self.engine._things_cache.overwriting(),
		):
			now = super()._delete(now=now)
			# don't advance time to store my non-location
			self.engine._things_cache.store(
				self.character.name, self.name, *now, ...
			)
			self.engine.query.set_thing_loc(
				self.character.name, self.name, *now, ...
			)
			return now
